@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2009 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -33,9 +33,9 @@
 	"[#{window_width}x#{window_height}] "			\
 	"(#{window_panes} panes) #{window_find_matches}"
 
-enum cmd_retval	 cmd_find_window_exec(struct cmd *, struct cmd_q *);
+static enum cmd_retval	cmd_find_window_exec(struct cmd *, struct cmdq_item *);
 
-void	cmd_find_window_callback(struct window_choose_data *);
+static void		cmd_find_window_callback(struct window_choose_data *);
 
 /* Flags for determining matching behavior. */
 #define CMD_FIND_WINDOW_BY_TITLE   0x1
@@ -48,11 +48,16 @@ void	cmd_find_window_callback(struct window_choose_data *);
 	 CMD_FIND_WINDOW_BY_NAME)
 
 const struct cmd_entry cmd_find_window_entry = {
-	"find-window", "findw",
-	"F:CNt:T", 1, 4,
-	"[-CNT] [-F format] " CMD_TARGET_WINDOW_USAGE " match-string",
-	0,
-	cmd_find_window_exec
+	.name = "find-window",
+	.alias = "findw",
+
+	.args = { "F:CNt:T", 1, 4 },
+	.usage = "[-CNT] [-F format] " CMD_TARGET_WINDOW_USAGE " match-string",
+
+	.target = { 't', CMD_FIND_WINDOW, 0 },
+
+	.flags = 0,
+	.exec = cmd_find_window_exec
 };
 
 struct cmd_find_window_data {
@@ -63,11 +68,11 @@ struct cmd_find_window_data {
 };
 TAILQ_HEAD(cmd_find_window_list, cmd_find_window_data);
 
-u_int	cmd_find_window_match_flags(struct args *);
-void	cmd_find_window_match(struct cmd_find_window_list *, int,
-	    struct winlink *, const char *, const char *);
+static u_int	cmd_find_window_match_flags(struct args *);
+static void	cmd_find_window_match(struct cmd_find_window_list *, int,
+		    struct winlink *, const char *, const char *);
 
-u_int
+static u_int
 cmd_find_window_match_flags(struct args *args)
 {
 	u_int	match_flags = 0;
@@ -87,7 +92,7 @@ cmd_find_window_match_flags(struct args *args)
 	return (match_flags);
 }
 
-void
+static void
 cmd_find_window_match(struct cmd_find_window_list *find_list,
     int match_flags, struct winlink *wl, const char *str,
     const char *searchstr)
@@ -133,14 +138,15 @@ cmd_find_window_match(struct cmd_find_window_list *find_list,
 		free(find_data);
 }
 
-enum cmd_retval
-cmd_find_window_exec(struct cmd *self, struct cmd_q *cmdq)
+static enum cmd_retval
+cmd_find_window_exec(struct cmd *self, struct cmdq_item *item)
 {
 	struct args			*args = self->args;
-	struct client			*c;
+	struct cmd_find_state		*current = &item->shared->current;
+	struct client			*c = cmd_find_client(item, NULL, 1);
 	struct window_choose_data	*cdata;
-	struct session			*s;
-	struct winlink			*wl, *wm;
+	struct session			*s = item->target.s;
+	struct winlink			*wl = item->target.wl, *wm;
 	struct cmd_find_window_list	 find_list;
 	struct cmd_find_window_data	*find_data;
 	struct cmd_find_window_data	*find_data1;
@@ -148,14 +154,10 @@ cmd_find_window_exec(struct cmd *self, struct cmd_q *cmdq)
 	const char			*template;
 	u_int				 i, match_flags;
 
-	if ((c = cmd_find_client(cmdq, NULL, 1)) == NULL) {
-		cmdq_error(cmdq, "no client available");
+	if (c == NULL) {
+		cmdq_error(item, "no client available");
 		return (CMD_RETURN_ERROR);
 	}
-	s = c->session;
-
-	if ((wl = cmd_find_window(cmdq, args_get(args, 't'), NULL)) == NULL)
-		return (CMD_RETURN_ERROR);
 
 	if ((template = args_get(args, 'F')) == NULL)
 		template = FIND_WINDOW_TEMPLATE;
@@ -171,13 +173,15 @@ cmd_find_window_exec(struct cmd *self, struct cmd_q *cmdq)
 	free(searchstr);
 
 	if (TAILQ_EMPTY(&find_list)) {
-		cmdq_error(cmdq, "no windows matching: %s", str);
+		cmdq_error(item, "no windows matching: %s", str);
 		return (CMD_RETURN_ERROR);
 	}
 
 	if (TAILQ_NEXT(TAILQ_FIRST(&find_list), entry) == NULL) {
-		if (session_select(s, TAILQ_FIRST(&find_list)->wl->idx) == 0)
+		if (session_select(s, TAILQ_FIRST(&find_list)->wl->idx) == 0) {
+			cmd_find_from_session(current, s);
 			server_redraw_session(s);
+		}
 		recalculate_sizes();
 		goto out;
 	}
@@ -215,7 +219,7 @@ out:
 	return (CMD_RETURN_NORMAL);
 }
 
-void
+static void
 cmd_find_window_callback(struct window_choose_data *cdata)
 {
 	struct session		*s;
