@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2015 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2015 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,66 +22,32 @@
 #include <limits.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "tmux.h"
 
-#define CMD_FIND_PREFER_UNATTACHED 0x1
-#define CMD_FIND_QUIET 0x2
-#define CMD_FIND_WINDOW_INDEX 0x4
-#define CMD_FIND_DEFAULT_MARKED 0x8
-#define CMD_FIND_EXACT_SESSION 0x10
-#define CMD_FIND_EXACT_WINDOW 0x20
+static int	cmd_find_session_better(struct session *, struct session *,
+		    int);
+static struct session *cmd_find_best_session(struct session **, u_int, int);
+static int	cmd_find_best_session_with_window(struct cmd_find_state *);
+static int	cmd_find_best_winlink_with_window(struct cmd_find_state *);
 
-enum cmd_find_type {
-	CMD_FIND_PANE,
-	CMD_FIND_WINDOW,
-	CMD_FIND_SESSION,
-};
+static const char *cmd_find_map_table(const char *[][2], const char *);
 
-struct cmd_find_state {
-	struct cmd_q		*cmdq;
-	int			 flags;
-	struct cmd_find_state	*current;
+static int	cmd_find_get_session(struct cmd_find_state *, const char *);
+static int	cmd_find_get_window(struct cmd_find_state *, const char *, int);
+static int	cmd_find_get_window_with_session(struct cmd_find_state *,
+		    const char *);
+static int	cmd_find_get_pane(struct cmd_find_state *, const char *, int);
+static int	cmd_find_get_pane_with_session(struct cmd_find_state *,
+		    const char *);
+static int	cmd_find_get_pane_with_window(struct cmd_find_state *,
+		    const char *);
 
-	struct session          *s;
-	struct winlink          *wl;
-	struct window		*w;
-	struct window_pane      *wp;
-	int			 idx;
-};
-
-int		 cmd_find_client_better(struct client *, struct client *);
-struct client	*cmd_find_best_client(struct client **, u_int);
-int		 cmd_find_session_better(struct session *, struct session *,
-		     int);
-struct session	*cmd_find_best_session(struct session **, u_int, int);
-int		 cmd_find_best_session_with_window(struct cmd_find_state *);
-int		 cmd_find_best_winlink_with_window(struct cmd_find_state *);
-
-int		 cmd_find_current_session_with_client(struct cmd_find_state *);
-int		 cmd_find_current_session(struct cmd_find_state *);
-struct client	*cmd_find_current_client(struct cmd_q *);
-
-const char	*cmd_find_map_table(const char *[][2], const char *);
-
-int	cmd_find_get_session(struct cmd_find_state *, const char *);
-int	cmd_find_get_window(struct cmd_find_state *, const char *);
-int	cmd_find_get_window_with_session(struct cmd_find_state *, const char *);
-int	cmd_find_get_window_with_pane(struct cmd_find_state *);
-int	cmd_find_get_pane(struct cmd_find_state *, const char *);
-int	cmd_find_get_pane_with_session(struct cmd_find_state *, const char *);
-int	cmd_find_get_pane_with_window(struct cmd_find_state *, const char *);
-
-void	cmd_find_clear_state(struct cmd_find_state *, struct cmd_q *, int);
-void	cmd_find_log_state(const char *, const char *, struct cmd_find_state *);
-
-struct cmd_find_state	*cmd_find_target(struct cmd_q *, const char *,
-	    enum cmd_find_type, int);
-
-const char *cmd_find_session_table[][2] = {
+static const char *cmd_find_session_table[][2] = {
 	{ NULL, NULL }
 };
-const char *cmd_find_window_table[][2] = {
+static const char *cmd_find_window_table[][2] = {
 	{ "{start}", "^" },
 	{ "{last}", "!" },
 	{ "{end}", "$" },
@@ -89,7 +55,7 @@ const char *cmd_find_window_table[][2] = {
 	{ "{previous}", "-" },
 	{ NULL, NULL }
 };
-const char *cmd_find_pane_table[][2] = {
+static const char *cmd_find_pane_table[][2] = {
 	{ "{last}", "!" },
 	{ "{next}", "+" },
 	{ "{previous}", "-" },
@@ -108,8 +74,29 @@ const char *cmd_find_pane_table[][2] = {
 	{ NULL, NULL }
 };
 
+/* Get session from TMUX if present. */
+static struct session *
+cmd_find_try_TMUX(struct client *c)
+{
+	struct environ_entry	*envent;
+	char			 tmp[256];
+	long long		 pid;
+	u_int			 session;
+
+	envent = environ_find(c->environ, "TMUX");
+	if (envent == NULL)
+		return (NULL);
+
+	if (sscanf(envent->value, "%255[^,],%lld,%d", tmp, &pid, &session) != 3)
+		return (NULL);
+	if (pid != getpid())
+		return (NULL);
+	log_debug("client %p TMUX %s (session @%u)", c, envent->value, session);
+	return (session_find_by_id(session));
+}
+
 /* Is this client better? */
-int
+static int
 cmd_find_client_better(struct client *c, struct client *than)
 {
 	if (than == NULL)
@@ -117,34 +104,29 @@ cmd_find_client_better(struct client *c, struct client *than)
 	return (timercmp(&c->activity_time, &than->activity_time, >));
 }
 
-/* Find best client from a list, or all if list is NULL. */
-struct client *
-cmd_find_best_client(struct client **clist, u_int csize)
+/* Find best client for session. */
+static struct client *
+cmd_find_best_client(struct session *s)
 {
 	struct client	*c_loop, *c;
-	u_int		 i;
+
+	if (s->flags & SESSION_UNATTACHED)
+		s = NULL;
 
 	c = NULL;
-	if (clist != NULL) {
-		for (i = 0; i < csize; i++) {
-			if (clist[i]->session == NULL)
-				continue;
-			if (cmd_find_client_better(clist[i], c))
-				c = clist[i];
-		}
-	} else {
-		TAILQ_FOREACH(c_loop, &clients, entry) {
-			if (c_loop->session == NULL)
-				continue;
-			if (cmd_find_client_better(c_loop, c))
-				c = c_loop;
-		}
+	TAILQ_FOREACH(c_loop, &clients, entry) {
+		if (c_loop->session == NULL)
+			continue;
+		if (s != NULL && c_loop->session != s)
+			continue;
+		if (cmd_find_client_better(c_loop, c))
+			c = c_loop;
 	}
 	return (c);
 }
 
 /* Is this session better? */
-int
+static int
 cmd_find_session_better(struct session *s, struct session *than, int flags)
 {
 	int	attached;
@@ -162,7 +144,7 @@ cmd_find_session_better(struct session *s, struct session *than, int flags)
 }
 
 /* Find best session from a list, or all if list is NULL. */
-struct session *
+static struct session *
 cmd_find_best_session(struct session **slist, u_int ssize, int flags)
 {
 	struct session	 *s_loop, *s;
@@ -184,7 +166,7 @@ cmd_find_best_session(struct session **slist, u_int ssize, int flags)
 }
 
 /* Find best session and winlink for window. */
-int
+static int
 cmd_find_best_session_with_window(struct cmd_find_state *fs)
 {
 	struct session	**slist = NULL;
@@ -215,13 +197,13 @@ fail:
  * Find the best winlink for a window (the current if it contains the pane,
  * otherwise the first).
  */
-int
+static int
 cmd_find_best_winlink_with_window(struct cmd_find_state *fs)
 {
 	struct winlink	 *wl, *wl_loop;
 
 	wl = NULL;
-	if (fs->s->curw->window == fs->w)
+	if (fs->s->curw != NULL && fs->s->curw->window == fs->w)
 		wl = fs->s->curw;
 	else {
 		RB_FOREACH(wl_loop, winlinks, &fs->s->windows) {
@@ -238,118 +220,8 @@ cmd_find_best_winlink_with_window(struct cmd_find_state *fs)
 	return (0);
 }
 
-/* Find current session when we have an unattached client. */
-int
-cmd_find_current_session_with_client(struct cmd_find_state *fs)
-{
-	struct window_pane	*wp;
-
-	/* If this is running in a pane, that's great. */
-	if (fs->cmdq->client->tty.path != NULL) {
-		RB_FOREACH(wp, window_pane_tree, &all_window_panes) {
-			if (strcmp(wp->tty, fs->cmdq->client->tty.path) == 0)
-				break;
-		}
-	} else
-		wp = NULL;
-
-	/* Not running in a pane. We know nothing. Find the best session. */
-	if (wp == NULL) {
-		fs->s = cmd_find_best_session(NULL, 0, fs->flags);
-		if (fs->s == NULL)
-			return (-1);
-		fs->wl = fs->s->curw;
-		fs->idx = fs->wl->idx;
-		fs->w = fs->wl->window;
-		fs->wp = fs->w->active;
-		return (0);
-	}
-
-	/* We now know the window and pane. */
-	fs->w = wp->window;
-	fs->wp = wp;
-
-	/* Find the best session and winlink. */
-	if (cmd_find_best_session_with_window(fs) != 0)
-		return (-1);
-	return (0);
-}
-
-/*
- * Work out the best current state. If this function succeeds, the state is
- * guaranteed to be completely filled in.
- */
-int
-cmd_find_current_session(struct cmd_find_state *fs)
-{
-	/* If we know the current client, use it. */
-	if (fs->cmdq->client != NULL) {
-		if (fs->cmdq->client->session == NULL)
-			return (cmd_find_current_session_with_client(fs));
-		fs->s = fs->cmdq->client->session;
-		fs->wl = fs->s->curw;
-		fs->idx = fs->wl->idx;
-		fs->w = fs->wl->window;
-		fs->wp = fs->w->active;
-		return (0);
-	}
-
-	/* We know nothing, find the best session and client. */
-	fs->s = cmd_find_best_session(NULL, 0, fs->flags);
-	if (fs->s == NULL)
-		return (-1);
-	fs->wl = fs->s->curw;
-	fs->idx = fs->wl->idx;
-	fs->w = fs->wl->window;
-	fs->wp = fs->w->active;
-
-	return (0);
-}
-
-/* Work out the best current client. */
-struct client *
-cmd_find_current_client(struct cmd_q *cmdq)
-{
-	struct cmd_find_state	 current;
-	struct session		*s;
-	struct client		*c, **clist = NULL;
-	u_int		 	 csize;
-
-	/* If the queue client has a session, use it. */
-	if (cmdq->client != NULL && cmdq->client->session != NULL)
-		return (cmdq->client);
-
-	/* Otherwise find the current session. */
-	cmd_find_clear_state(&current, cmdq, 0);
-	if (cmd_find_current_session(&current) != 0)
-		return (NULL);
-
-	/* If it is attached, find the best of it's clients. */
-	s = current.s;
-	if (~s->flags & SESSION_UNATTACHED) {
-		csize = 0;
-		TAILQ_FOREACH(c, &clients, entry) {
-			if (c->session != s)
-				continue;
-			clist = xreallocarray(clist, csize + 1, sizeof *clist);
-			clist[csize++] = c;
-		}
-		if (csize != 0) {
-			c = cmd_find_best_client(clist, csize);
-			if (c != NULL) {
-				free(clist);
-				return (c);
-			}
-		}
-		free(clist);
-	}
-
-	/* Otherwise pick best of all clients. */
-	return (cmd_find_best_client(NULL, 0));
-}
-
 /* Maps string in table. */
-const char *
+static const char *
 cmd_find_map_table(const char *table[][2], const char *s)
 {
 	u_int	i;
@@ -362,10 +234,11 @@ cmd_find_map_table(const char *table[][2], const char *s)
 }
 
 /* Find session from string. Fills in s. */
-int
+static int
 cmd_find_get_session(struct cmd_find_state *fs, const char *session)
 {
 	struct session	*s, *s_loop;
+	struct client	*c;
 
 	log_debug("%s: %s", __func__, session);
 
@@ -382,6 +255,13 @@ cmd_find_get_session(struct cmd_find_state *fs, const char *session)
 	if (fs->s != NULL)
 		return (0);
 
+	/* Look for as a client. */
+	c = cmd_find_client(NULL, session, 1);
+	if (c != NULL && c->session != NULL) {
+		fs->s = c->session;
+		return (0);
+	}
+
 	/* Stop now if exact only. */
 	if (fs->flags & CMD_FIND_EXACT_SESSION)
 		return (-1);
@@ -418,8 +298,8 @@ cmd_find_get_session(struct cmd_find_state *fs, const char *session)
 }
 
 /* Find window from string. Fills in s, wl, w. */
-int
-cmd_find_get_window(struct cmd_find_state *fs, const char *window)
+static int
+cmd_find_get_window(struct cmd_find_state *fs, const char *window, int only)
 {
 	log_debug("%s: %s", __func__, window);
 
@@ -439,12 +319,11 @@ cmd_find_get_window(struct cmd_find_state *fs, const char *window)
 		return (0);
 
 	/* Otherwise try as a session itself. */
-	if (cmd_find_get_session(fs, window) == 0) {
-		if (~fs->flags & CMD_FIND_WINDOW_INDEX) {
-			fs->wl = fs->s->curw;
-			fs->w = fs->wl->window;
+	if (!only && cmd_find_get_session(fs, window) == 0) {
+		fs->wl = fs->s->curw;
+		fs->w = fs->wl->window;
+		if (~fs->flags & CMD_FIND_WINDOW_INDEX)
 			fs->idx = fs->wl->idx;
-		}
 		return (0);
 	}
 
@@ -455,7 +334,7 @@ cmd_find_get_window(struct cmd_find_state *fs, const char *window)
  * Find window from string, assuming it is in given session. Needs s, fills in
  * wl and w.
  */
-int
+static int
 cmd_find_get_window_with_session(struct cmd_find_state *fs, const char *window)
 {
 	struct winlink	*wl;
@@ -466,6 +345,13 @@ cmd_find_get_window_with_session(struct cmd_find_state *fs, const char *window)
 	log_debug("%s: %s", __func__, window);
 	exact = (fs->flags & CMD_FIND_EXACT_WINDOW);
 
+	/*
+	 * Start with the current window as the default. So if only an index is
+	 * found, the window will be the current.
+	 */
+	fs->wl = fs->s->curw;
+	fs->w = fs->wl->window;
+
 	/* Check for window ids starting with @. */
 	if (*window == '@') {
 		fs->w = window_find_by_id_str(window);
@@ -598,26 +484,16 @@ cmd_find_get_window_with_session(struct cmd_find_state *fs, const char *window)
 	return (-1);
 }
 
-/* Find window from given pane. Needs wp, fills in s and wl and w. */
-int
-cmd_find_get_window_with_pane(struct cmd_find_state *fs)
-{
-	log_debug("%s", __func__);
-
-	fs->w = fs->wp->window;
-	return (cmd_find_best_session_with_window(fs));
-}
-
 /* Find pane from string. Fills in s, wl, w, wp. */
-int
-cmd_find_get_pane(struct cmd_find_state *fs, const char *pane)
+static int
+cmd_find_get_pane(struct cmd_find_state *fs, const char *pane, int only)
 {
 	log_debug("%s: %s", __func__, pane);
 
 	/* Check for pane ids starting with %. */
 	if (*pane == '%') {
 		fs->wp = window_pane_find_by_id_str(pane);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		fs->w = fs->wp->window;
 		return (cmd_find_best_session_with_window(fs));
@@ -634,7 +510,7 @@ cmd_find_get_pane(struct cmd_find_state *fs, const char *pane)
 		return (0);
 
 	/* Otherwise try as a window itself (this will also try as session). */
-	if (cmd_find_get_window(fs, pane) == 0) {
+	if (!only && cmd_find_get_window(fs, pane, 0) == 0) {
 		fs->wp = fs->w->active;
 		return (0);
 	}
@@ -646,7 +522,7 @@ cmd_find_get_pane(struct cmd_find_state *fs, const char *pane)
  * Find pane from string, assuming it is in given session. Needs s, fills in wl
  * and w and wp.
  */
-int
+static int
 cmd_find_get_pane_with_session(struct cmd_find_state *fs, const char *pane)
 {
 	log_debug("%s: %s", __func__, pane);
@@ -654,7 +530,7 @@ cmd_find_get_pane_with_session(struct cmd_find_state *fs, const char *pane)
 	/* Check for pane ids starting with %. */
 	if (*pane == '%') {
 		fs->wp = window_pane_find_by_id_str(pane);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		fs->w = fs->wp->window;
 		return (cmd_find_best_winlink_with_window(fs));
@@ -673,7 +549,7 @@ cmd_find_get_pane_with_session(struct cmd_find_state *fs, const char *pane)
  * Find pane from string, assuming it is in the given window. Needs w, fills in
  * wp.
  */
-int
+static int
 cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 {
 	const char		*errstr;
@@ -686,7 +562,9 @@ cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 	/* Check for pane ids starting with %. */
 	if (*pane == '%') {
 		fs->wp = window_pane_find_by_id_str(pane);
-		if (fs->wp == NULL || fs->wp->window != fs->w)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
+			return (-1);
+		if (fs->wp->window != fs->w)
 			return (-1);
 		return (0);
 	}
@@ -696,25 +574,27 @@ cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 		if (fs->w->last == NULL)
 			return (-1);
 		fs->wp = fs->w->last;
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
+			return (-1);
 		return (0);
 	} else if (strcmp(pane, "{up-of}") == 0) {
 		fs->wp = window_pane_find_up(fs->w->active);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		return (0);
 	} else if (strcmp(pane, "{down-of}") == 0) {
 		fs->wp = window_pane_find_down(fs->w->active);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		return (0);
 	} else if (strcmp(pane, "{left-of}") == 0) {
 		fs->wp = window_pane_find_left(fs->w->active);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		return (0);
 	} else if (strcmp(pane, "{right-of}") == 0) {
 		fs->wp = window_pane_find_right(fs->w->active);
-		if (fs->wp == NULL)
+		if (fs->wp == NULL || window_pane_outside(fs->wp))
 			return (-1);
 		return (0);
 	}
@@ -730,7 +610,7 @@ cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 			fs->wp = window_pane_next_by_number(fs->w, wp, n);
 		else
 			fs->wp = window_pane_previous_by_number(fs->w, wp, n);
-		if (fs->wp != NULL)
+		if (fs->wp != NULL && !window_pane_outside(fs->wp))
 			return (0);
 	}
 
@@ -738,13 +618,13 @@ cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 	idx = strtonum(pane, 0, INT_MAX, &errstr);
 	if (errstr == NULL) {
 		fs->wp = window_pane_at_index(fs->w, idx);
-		if (fs->wp != NULL)
+		if (fs->wp != NULL && !window_pane_outside(fs->wp))
 			return (0);
 	}
 
 	/* Try as a description. */
 	fs->wp = window_find_string(fs->w, pane);
-	if (fs->wp != NULL)
+	if (fs->wp != NULL && !window_pane_outside(fs->wp))
 		return (0);
 
 	return (-1);
@@ -752,44 +632,354 @@ cmd_find_get_pane_with_window(struct cmd_find_state *fs, const char *pane)
 
 /* Clear state. */
 void
-cmd_find_clear_state(struct cmd_find_state *fs, struct cmd_q *cmdq, int flags)
+cmd_find_clear_state(struct cmd_find_state *fs, int flags)
 {
 	memset(fs, 0, sizeof *fs);
 
-	fs->cmdq = cmdq;
 	fs->flags = flags;
 
 	fs->idx = -1;
 }
 
-/* Split target into pieces and resolve for the given type. */
-struct cmd_find_state *
-cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
-    int flags)
+/* Check if state is empty. */
+int
+cmd_find_empty_state(struct cmd_find_state *fs)
 {
-	static struct cmd_find_state	 fs, current;
-	struct mouse_event		*m;
-	char				*colon, *period, *copy = NULL;
-	const char			*session, *window, *pane;
+	if (fs->s == NULL && fs->wl == NULL && fs->w == NULL && fs->wp == NULL)
+		return (1);
+	return (0);
+}
 
-	/* Find current state. */
-	cmd_find_clear_state(&current, cmdq, flags);
-	if (server_check_marked() && (flags & CMD_FIND_DEFAULT_MARKED)) {
-		current.s = marked_session;
-		current.wl = marked_winlink;
-		current.idx = current.wl->idx;
-		current.w = current.wl->window;
-		current.wp = marked_window_pane;
+/* Check if a state if valid. */
+int
+cmd_find_valid_state(struct cmd_find_state *fs)
+{
+	struct winlink	*wl;
+
+	if (fs->s == NULL || fs->wl == NULL || fs->w == NULL || fs->wp == NULL)
+		return (0);
+
+	if (!session_alive(fs->s))
+		return (0);
+
+	RB_FOREACH(wl, winlinks, &fs->s->windows) {
+		if (wl->window == fs->w && wl == fs->wl)
+			break;
 	}
-	if (current.s == NULL && cmd_find_current_session(&current) != 0) {
-		if (~flags & CMD_FIND_QUIET)
-			cmdq_error(cmdq, "no current session");
-		goto error;
+	if (wl == NULL)
+		return (0);
+
+	if (fs->w != fs->wl->window)
+		return (0);
+
+	if (!window_has_pane(fs->w, fs->wp))
+		return (0);
+	return (!window_pane_outside(fs->wp));
+}
+
+/* Copy a state. */
+void
+cmd_find_copy_state(struct cmd_find_state *dst, struct cmd_find_state *src)
+{
+	dst->s = src->s;
+	dst->wl = src->wl;
+	dst->idx = src->idx;
+	dst->w = src->w;
+	dst->wp = src->wp;
+}
+
+/* Log the result. */
+void
+cmd_find_log_state(const char *prefix, struct cmd_find_state *fs)
+{
+	if (fs->s != NULL)
+		log_debug("%s: s=$%u", prefix, fs->s->id);
+	else
+		log_debug("%s: s=none", prefix);
+	if (fs->wl != NULL) {
+		log_debug("%s: wl=%u %d w=@%u %s", prefix, fs->wl->idx,
+		    fs->wl->window == fs->w, fs->w->id, fs->w->name);
+	} else
+		log_debug("%s: wl=none", prefix);
+	if (fs->wp != NULL)
+		log_debug("%s: wp=%%%u", prefix, fs->wp->id);
+	else
+		log_debug("%s: wp=none", prefix);
+	if (fs->idx != -1)
+		log_debug("%s: idx=%d", prefix, fs->idx);
+	else
+		log_debug("%s: idx=none", prefix);
+}
+
+/* Find state from a session. */
+void
+cmd_find_from_session(struct cmd_find_state *fs, struct session *s)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->s = s;
+	fs->wl = fs->s->curw;
+	fs->w = fs->wl->window;
+	fs->wp = fs->w->active;
+
+	cmd_find_log_state(__func__, fs);
+}
+
+/* Find state from a winlink. */
+void
+cmd_find_from_winlink(struct cmd_find_state *fs, struct winlink *wl)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->s = wl->session;
+	fs->wl = wl;
+	fs->w = wl->window;
+	fs->wp = wl->window->active;
+
+	cmd_find_log_state(__func__, fs);
+}
+
+/* Find state from a session and window. */
+int
+cmd_find_from_session_window(struct cmd_find_state *fs, struct session *s,
+    struct window *w)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->s = s;
+	fs->w = w;
+	if (cmd_find_best_winlink_with_window(fs) != 0) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
 	}
+	fs->wp = fs->w->active;
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+}
+
+/* Find state from a window. */
+int
+cmd_find_from_window(struct cmd_find_state *fs, struct window *w)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->w = w;
+	if (cmd_find_best_session_with_window(fs) != 0) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	if (cmd_find_best_winlink_with_window(fs) != 0) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	fs->wp = fs->w->active;
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+}
+
+/* Find state from a winlink and pane. */
+void
+cmd_find_from_winlink_pane(struct cmd_find_state *fs, struct winlink *wl,
+    struct window_pane *wp)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->s = wl->session;
+	fs->wl = wl;
+	fs->idx = fs->wl->idx;
+	fs->w = fs->wl->window;
+	fs->wp = wp;
+
+	cmd_find_log_state(__func__, fs);
+}
+
+/* Find state from a pane. */
+int
+cmd_find_from_pane(struct cmd_find_state *fs, struct window_pane *wp)
+{
+	if (cmd_find_from_window(fs, wp->window) != 0)
+		return (-1);
+	if (window_pane_outside(wp)) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	fs->wp = wp;
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+}
+
+/* Find state from nothing. */
+int
+cmd_find_from_nothing(struct cmd_find_state *fs)
+{
+	cmd_find_clear_state(fs, 0);
+
+	fs->s = cmd_find_best_session(NULL, 0, fs->flags);
+	if (fs->s == NULL) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	fs->wl = fs->s->curw;
+	fs->idx = fs->wl->idx;
+	fs->w = fs->wl->window;
+	fs->wp = fs->w->active;
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+}
+
+/* Find state from mouse. */
+int
+cmd_find_from_mouse(struct cmd_find_state *fs, struct mouse_event *m)
+{
+	cmd_find_clear_state(fs, 0);
+
+	if (!m->valid)
+		return (-1);
+
+	fs->wp = cmd_mouse_pane(m, &fs->s, &fs->wl);
+	if (fs->wp == NULL) {
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	fs->w = fs->wl->window;
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+}
+
+/* Find state from client. */
+int
+cmd_find_from_client(struct cmd_find_state *fs, struct client *c)
+{
+	struct session		*s;
+	struct winlink		*wl;
+	struct window_pane	*wp;
+
+	/* If no client, treat as from nothing. */
+	if (c == NULL)
+		return (cmd_find_from_nothing(fs));
+
+	/* If this is an attached client, all done. */
+	if (c->session != NULL) {
+		cmd_find_from_session(fs, c->session);
+		return (0);
+	}
+	cmd_find_clear_state(fs, 0);
+
+	/*
+	 * If this is an unattached client running in a pane, we can use that
+	 * to limit the list of sessions to those containing that pane.
+	 */
+	RB_FOREACH(wp, window_pane_tree, &all_window_panes) {
+		if (strcmp(wp->tty, c->ttyname) == 0)
+			break;
+	}
+	if (wp == NULL)
+		goto unknown_pane;
+
+	/* If we have a session in TMUX, see if it has this pane. */
+	s = cmd_find_try_TMUX(c);
+	if (s != NULL) {
+		RB_FOREACH(wl, winlinks, &s->windows) {
+			if (window_has_pane(wl->window, wp))
+				break;
+		}
+		if (wl != NULL) {
+			fs->s = s;
+			fs->wl = s->curw; /* use current session */
+			fs->w = fs->wl->window;
+			fs->wp = fs->w->active; /* use active pane */
+
+			cmd_find_log_state(__func__, fs);
+			return (0);
+		}
+	}
+
+	/*
+	 * Don't have a session, or it doesn't have this pane. Try all
+	 * sessions.
+	 */
+	fs->w = wp->window;
+	if (cmd_find_best_session_with_window(fs) != 0) {
+		if (wp != NULL) {
+			/*
+			 * The window may have been destroyed but the pane
+			 * still on all_window_panes due to something else
+			 * holding a reference.
+			 */
+			goto unknown_pane;
+		}
+		cmd_find_clear_state(fs, 0);
+		return (-1);
+	}
+	fs->wl = fs->s->curw;
+	fs->w = fs->wl->window;
+	fs->wp = fs->w->active; /* use active pane */
+
+	cmd_find_log_state(__func__, fs);
+	return (0);
+
+unknown_pane:
+	/*
+	 * We're not running in a known pane, but maybe this client has TMUX
+	 * in the environment. That'd give us a session.
+	 */
+	s = cmd_find_try_TMUX(c);
+	if (s != NULL) {
+		cmd_find_from_session(fs, s);
+		return (0);
+	}
+
+	/* Otherwise we need to guess. */
+	return (cmd_find_from_nothing(fs));
+}
+
+/*
+ * Split target into pieces and resolve for the given type. Fills in the given
+ * state. Returns 0 on success or -1 on error.
+ */
+int
+cmd_find_target(struct cmd_find_state *fs, struct cmdq_item *item,
+    const char *target, enum cmd_find_type type, int flags)
+{
+	struct mouse_event	*m;
+	struct cmd_find_state	 current;
+	char			*colon, *period, *copy = NULL;
+	const char		*session, *window, *pane;
+	int			 window_only = 0, pane_only = 0;
+
+	/* Can fail flag implies quiet. */
+	if (flags & CMD_FIND_CANFAIL)
+		flags |= CMD_FIND_QUIET;
+
+	/* Log the arguments. */
+	if (target == NULL)
+		log_debug("%s: target none, type %d", __func__, type);
+	else
+		log_debug("%s: target %s, type %d", __func__, target, type);
+	log_debug("%s: item %p, flags %#x", __func__, item, flags);
 
 	/* Clear new state. */
-	cmd_find_clear_state(&fs, cmdq, flags);
-	fs.current = &current;
+	cmd_find_clear_state(fs, flags);
+
+	/* Find current state. */
+	if (server_check_marked() && (flags & CMD_FIND_DEFAULT_MARKED)) {
+		fs->current = &marked_pane;
+		log_debug("%s: current is marked pane", __func__);
+	} else if (cmd_find_valid_state(&item->shared->current)) {
+		fs->current = &item->shared->current;
+		log_debug("%s: current is from queue", __func__);
+	} else if (cmd_find_from_client(&current, item->client) == 0) {
+		fs->current = &current;
+		log_debug("%s: current is from client", __func__);
+	} else
+		goto error;
+	if (!cmd_find_valid_state(fs->current))
+		fatalx("invalid current find state");
 
 	/* An empty or NULL target is the current. */
 	if (target == NULL || *target == '\0')
@@ -797,43 +987,39 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 
 	/* Mouse target is a plain = or {mouse}. */
 	if (strcmp(target, "=") == 0 || strcmp(target, "{mouse}") == 0) {
-		m = &cmdq->item->mouse;
+		m = &item->shared->mouse;
 		switch (type) {
 		case CMD_FIND_PANE:
-			fs.wp = cmd_mouse_pane(m, &fs.s, &fs.wl);
-			if (fs.wp != NULL)
-				fs.w = fs.wl->window;
+			fs->wp = cmd_mouse_pane(m, &fs->s, &fs->wl);
+			if (fs->wp != NULL && !window_pane_outside(fs->wp))
+				fs->w = fs->wl->window;
 			break;
 		case CMD_FIND_WINDOW:
 		case CMD_FIND_SESSION:
-			fs.wl = cmd_mouse_window(m, &fs.s);
-			if (fs.wl != NULL) {
-				fs.w = fs.wl->window;
-				fs.wp = fs.w->active;
+			fs->wl = cmd_mouse_window(m, &fs->s);
+			if (fs->wl != NULL) {
+				fs->w = fs->wl->window;
+				fs->wp = fs->w->active;
 			}
 			break;
 		}
-		if (fs.wp == NULL) {
+		if (fs->wp == NULL) {
 			if (~flags & CMD_FIND_QUIET)
-				cmdq_error(cmdq, "no mouse target");
+				cmdq_error(item, "no mouse target");
 			goto error;
 		}
-		return (&fs);
+		goto found;
 	}
 
 	/* Marked target is a plain ~ or {marked}. */
 	if (strcmp(target, "~") == 0 || strcmp(target, "{marked}") == 0) {
 		if (!server_check_marked()) {
 			if (~flags & CMD_FIND_QUIET)
-				cmdq_error(cmdq, "no marked target");
+				cmdq_error(item, "no marked target");
 			goto error;
 		}
-		fs.s = marked_session;
-		fs.wl = marked_winlink;
-		fs.idx = fs.wl->idx;
-		fs.w = fs.wl->window;
-		fs.wp = marked_window_pane;
-		return (&fs);
+		cmd_find_copy_state(fs, &marked_pane);
+		goto found;
 	}
 
 	/* Find separators if they exist. */
@@ -853,13 +1039,17 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 	if (colon != NULL && period != NULL) {
 		session = copy;
 		window = colon;
+		window_only = 1;
 		pane = period;
+		pane_only = 1;
 	} else if (colon != NULL && period == NULL) {
 		session = copy;
 		window = colon;
+		window_only = 1;
 	} else if (colon == NULL && period != NULL) {
 		window = copy;
 		pane = period;
+		pane_only = 1;
 	} else {
 		if (*copy == '$')
 			session = copy;
@@ -885,11 +1075,11 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 	/* Set exact match flags. */
 	if (session != NULL && *session == '=') {
 		session++;
-		fs.flags |= CMD_FIND_EXACT_SESSION;
+		fs->flags |= CMD_FIND_EXACT_SESSION;
 	}
 	if (window != NULL && *window == '=') {
 		window++;
-		fs.flags |= CMD_FIND_EXACT_WINDOW;
+		fs->flags |= CMD_FIND_EXACT_WINDOW;
 	}
 
 	/* Empty is the same as NULL. */
@@ -908,46 +1098,45 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 	if (pane != NULL)
 		pane = cmd_find_map_table(cmd_find_pane_table, pane);
 
-	log_debug("target %s (flags %#x): session=%s, window=%s, pane=%s",
-	    target, flags, session == NULL ? "none" : session,
+	log_debug("%s: target %s (flags %#x): session=%s, window=%s, pane=%s",
+	    __func__, target, flags, session == NULL ? "none" : session,
 	    window == NULL ? "none" : window, pane == NULL ? "none" : pane);
 
 	/* No pane is allowed if want an index. */
 	if (pane != NULL && (flags & CMD_FIND_WINDOW_INDEX)) {
 		if (~flags & CMD_FIND_QUIET)
-			cmdq_error(cmdq, "can't specify pane here");
+			cmdq_error(item, "can't specify pane here");
 		goto error;
 	}
 
 	/* If the session isn't NULL, look it up. */
 	if (session != NULL) {
 		/* This will fill in session. */
-		if (cmd_find_get_session(&fs, session) != 0)
+		if (cmd_find_get_session(fs, session) != 0)
 			goto no_session;
 
 		/* If window and pane are NULL, use that session's current. */
 		if (window == NULL && pane == NULL) {
-			fs.wl = fs.s->curw;
-			fs.idx = -1;
-			fs.w = fs.wl->window;
-			fs.wp = fs.w->active;
+			fs->wl = fs->s->curw;
+			fs->idx = -1;
+			fs->w = fs->wl->window;
+			fs->wp = fs->w->active;
 			goto found;
 		}
 
 		/* If window is present but pane not, find window in session. */
 		if (window != NULL && pane == NULL) {
 			/* This will fill in winlink and window. */
-			if (cmd_find_get_window_with_session(&fs, window) != 0)
+			if (cmd_find_get_window_with_session(fs, window) != 0)
 				goto no_window;
-			if (~flags & CMD_FIND_WINDOW_INDEX)
-				fs.wp = fs.wl->window->active;
+			fs->wp = fs->wl->window->active;
 			goto found;
 		}
 
 		/* If pane is present but window not, find pane. */
 		if (window == NULL && pane != NULL) {
 			/* This will fill in winlink and window and pane. */
-			if (cmd_find_get_pane_with_session(&fs, pane) != 0)
+			if (cmd_find_get_pane_with_session(fs, pane) != 0)
 				goto no_pane;
 			goto found;
 		}
@@ -956,10 +1145,10 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 		 * If window and pane are present, find both in session. This
 		 * will fill in winlink and window.
 		 */
-		if (cmd_find_get_window_with_session(&fs, window) != 0)
+		if (cmd_find_get_window_with_session(fs, window) != 0)
 			goto no_window;
 		/* This will fill in pane. */
-		if (cmd_find_get_pane_with_window(&fs, pane) != 0)
+		if (cmd_find_get_pane_with_window(fs, pane) != 0)
 			goto no_pane;
 		goto found;
 	}
@@ -967,10 +1156,10 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 	/* No session. If window and pane, try them. */
 	if (window != NULL && pane != NULL) {
 		/* This will fill in session, winlink and window. */
-		if (cmd_find_get_window(&fs, window) != 0)
+		if (cmd_find_get_window(fs, window, window_only) != 0)
 			goto no_window;
 		/* This will fill in pane. */
-		if (cmd_find_get_pane_with_window(&fs, pane) != 0)
+		if (cmd_find_get_pane_with_window(fs, pane) != 0)
 			goto no_pane;
 		goto found;
 	}
@@ -978,199 +1167,81 @@ cmd_find_target(struct cmd_q *cmdq, const char *target, enum cmd_find_type type,
 	/* If just window is present, try it. */
 	if (window != NULL && pane == NULL) {
 		/* This will fill in session, winlink and window. */
-		if (cmd_find_get_window(&fs, window) != 0)
+		if (cmd_find_get_window(fs, window, window_only) != 0)
 			goto no_window;
-		if (~flags & CMD_FIND_WINDOW_INDEX)
-			fs.wp = fs.wl->window->active;
+		fs->wp = fs->wl->window->active;
 		goto found;
 	}
 
 	/* If just pane is present, try it. */
 	if (window == NULL && pane != NULL) {
 		/* This will fill in session, winlink, window and pane. */
-		if (cmd_find_get_pane(&fs, pane) != 0)
+		if (cmd_find_get_pane(fs, pane, pane_only) != 0)
 			goto no_pane;
 		goto found;
 	}
 
 current:
-	/* None is the current session. */
-	free(copy);
+	/* Use the current session. */
+	cmd_find_copy_state(fs, fs->current);
 	if (flags & CMD_FIND_WINDOW_INDEX)
-		current.idx = -1;
-	return (&current);
+		fs->idx = -1;
+	goto found;
 
 error:
+	fs->current = NULL;
+	log_debug("%s: error", __func__);
+
 	free(copy);
-	return (NULL);
+	if (flags & CMD_FIND_CANFAIL)
+		return (0);
+	return (-1);
 
 found:
+	fs->current = NULL;
+	cmd_find_log_state(__func__, fs);
+
 	free(copy);
-	return (&fs);
+	return (0);
 
 no_session:
 	if (~flags & CMD_FIND_QUIET)
-		cmdq_error(cmdq, "can't find session %s", session);
+		cmdq_error(item, "can't find session %s", session);
 	goto error;
 
 no_window:
 	if (~flags & CMD_FIND_QUIET)
-		cmdq_error(cmdq, "can't find window %s", window);
+		cmdq_error(item, "can't find window %s", window);
 	goto error;
 
 no_pane:
 	if (~flags & CMD_FIND_QUIET)
-		cmdq_error(cmdq, "can't find pane %s", pane);
+		cmdq_error(item, "can't find pane %s", pane);
 	goto error;
 }
 
-/* Log the result. */
-void
-cmd_find_log_state(const char *f, const char *target, struct cmd_find_state *fs)
-{
-	log_debug("%s: target %s%s", f, target == NULL ? "none" : target,
-	    fs != NULL ? "" : " (failed)");
-	if (fs == NULL)
-		return;
-	if (fs->s != NULL)
-		log_debug("\ts=$%u", fs->s->id);
-	else
-		log_debug("\ts=none");
-	if (fs->wl != NULL) {
-		log_debug("\twl=%u %d w=@%u %s", fs->wl->idx,
-		    fs->wl->window == fs->w, fs->w->id, fs->w->name);
-	} else
-		log_debug("\twl=none");
-	if (fs->wp != NULL)
-		log_debug("\twp=%%%u", fs->wp->id);
-	else
-		log_debug("\twp=none");
-	if (fs->idx != -1)
-		log_debug("\tidx=%d", fs->idx);
-	else
-		log_debug("\tidx=none");
-}
-
-/* Find the current session. */
-struct session *
-cmd_find_current(struct cmd_q *cmdq)
-{
-	struct cmd_find_state	*fs;
-	int			 flags = CMD_FIND_QUIET;
-
-	fs = cmd_find_target(cmdq, NULL, CMD_FIND_SESSION, flags);
-	cmd_find_log_state(__func__, NULL, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	return (fs->s);
-}
-
-/* Find the target session or report an error and return NULL. */
-struct session *
-cmd_find_session(struct cmd_q *cmdq, const char *target, int prefer_unattached)
-{
-	struct cmd_find_state	*fs;
-	int			 flags = 0;
-
-	if (prefer_unattached)
-		flags |= CMD_FIND_PREFER_UNATTACHED;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_SESSION, flags);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	return (fs->s);
-}
-
-/* Find the target window or report an error and return NULL. */
-struct winlink *
-cmd_find_window(struct cmd_q *cmdq, const char *target, struct session **sp)
-{
-	struct cmd_find_state	*fs;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_WINDOW, 0);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	if (sp != NULL)
-		*sp = fs->s;
-	return (fs->wl);
-}
-
-/* Find the target window, defaulting to marked rather than current. */
-struct winlink *
-cmd_find_window_marked(struct cmd_q *cmdq, const char *target,
-    struct session **sp)
-{
-	struct cmd_find_state	*fs;
-	int			 flags = CMD_FIND_DEFAULT_MARKED;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_WINDOW, flags);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	if (sp != NULL)
-		*sp = fs->s;
-	return (fs->wl);
-}
-
-/* Find the target pane and report an error and return NULL. */
-struct winlink *
-cmd_find_pane(struct cmd_q *cmdq, const char *target, struct session **sp,
-    struct window_pane **wpp)
-{
-	struct cmd_find_state	*fs;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_PANE, 0);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	if (sp != NULL)
-		*sp = fs->s;
-	if (wpp != NULL)
-		*wpp = fs->wp;
-	return (fs->wl);
-}
-
-/* Find the target pane, defaulting to marked rather than current. */
-struct winlink *
-cmd_find_pane_marked(struct cmd_q *cmdq, const char *target,
-    struct session **sp, struct window_pane **wpp)
-{
-	struct cmd_find_state	*fs;
-	int			 flags = CMD_FIND_DEFAULT_MARKED;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_PANE, flags);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (NULL);
-
-	if (sp != NULL)
-		*sp = fs->s;
-	if (wpp != NULL)
-		*wpp = fs->wp;
-	return (fs->wl);
-}
-
 /* Find the target client or report an error and return NULL. */
 struct client *
-cmd_find_client(struct cmd_q *cmdq, const char *target, int quiet)
+cmd_find_client(struct cmdq_item *item, const char *target, int quiet)
 {
 	struct client	*c;
+	struct session	*s;
 	char		*copy;
 	size_t		 size;
-	const char	*path;
 
 	/* A NULL argument means the current client. */
 	if (target == NULL) {
-		c = cmd_find_current_client(cmdq);
+		c = NULL;
+		if (item->client != NULL && item->client->session != NULL)
+			c = item->client;
+		else {
+			s = cmd_find_best_session(NULL, 0, CMD_FIND_QUIET);
+			if (s != NULL)
+				c = cmd_find_best_client(s);
+		}
 		if (c == NULL && !quiet)
-			cmdq_error(cmdq, "no current client");
+			cmdq_error(item, "no current client");
+		log_debug("%s: no target, return %p", __func__, c);
 		return (c);
 	}
 	copy = xstrdup(target);
@@ -1180,49 +1251,28 @@ cmd_find_client(struct cmd_q *cmdq, const char *target, int quiet)
 	if (size != 0 && copy[size - 1] == ':')
 		copy[size - 1] = '\0';
 
-	/* Check path of each client. */
+	/* Check name and path of each client. */
 	TAILQ_FOREACH(c, &clients, entry) {
-		if (c->session == NULL || c->tty.path == NULL)
+		if (c->session == NULL)
 			continue;
-		path = c->tty.path;
-
-		/* Try for exact match. */
-		if (strcmp(copy, path) == 0)
+		if (strcmp(copy, c->name) == 0)
 			break;
 
-		/* Try without leading /dev. */
-		if (strncmp(path, _PATH_DEV, (sizeof _PATH_DEV) - 1) != 0)
+		if (*c->ttyname == '\0')
 			continue;
-		if (strcmp(copy, path + (sizeof _PATH_DEV) - 1) == 0)
+		if (strcmp(copy, c->ttyname) == 0)
+			break;
+		if (strncmp(c->ttyname, _PATH_DEV, (sizeof _PATH_DEV) - 1) != 0)
+			continue;
+		if (strcmp(copy, c->ttyname + (sizeof _PATH_DEV) - 1) == 0)
 			break;
 	}
 
 	/* If no client found, report an error. */
 	if (c == NULL && !quiet)
-		cmdq_error(cmdq, "can't find client %s", copy);
+		cmdq_error(item, "can't find client %s", copy);
 
 	free(copy);
+	log_debug("%s: target %s, return %p", __func__, target, c);
 	return (c);
 }
-
-/*
- * Find the target session and window index, whether or not it exists in the
- * session. Return -2 on error or -1 if no window index is specified. This is
- * used when parsing an argument for a window target that may not exist (for
- * example if it is going to be created).
- */
-int
-cmd_find_index(struct cmd_q *cmdq, const char *target, struct session **sp)
-{
-	struct cmd_find_state	*fs;
-	int			 flags = CMD_FIND_WINDOW_INDEX;
-
-	fs = cmd_find_target(cmdq, target, CMD_FIND_WINDOW, flags);
-	cmd_find_log_state(__func__, target, fs);
-	if (fs == NULL)
-		return (-2);
-
-	if (sp != NULL)
-		*sp = fs->s;
-	return (fs->idx);
-}
