@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -27,26 +27,37 @@
  * Note this deliberately has no alias to make it hard to hit by accident.
  */
 
-enum cmd_retval	 cmd_kill_session_exec(struct cmd *, struct cmd_q *);
+static enum cmd_retval	cmd_kill_session_exec(struct cmd *, struct cmdq_item *);
 
 const struct cmd_entry cmd_kill_session_entry = {
-	"kill-session", NULL,
-	"at:", 0, 0,
-	"[-a] " CMD_TARGET_SESSION_USAGE,
-	0,
-	cmd_kill_session_exec
+	.name = "kill-session",
+	.alias = NULL,
+
+	.args = { "aCt:", 0, 0 },
+	.usage = "[-aC] " CMD_TARGET_SESSION_USAGE,
+
+	.target = { 't', CMD_FIND_SESSION, 0 },
+
+	.flags = 0,
+	.exec = cmd_kill_session_exec
 };
 
-enum cmd_retval
-cmd_kill_session_exec(struct cmd *self, struct cmd_q *cmdq)
+static enum cmd_retval
+cmd_kill_session_exec(struct cmd *self, struct cmdq_item *item)
 {
 	struct args	*args = self->args;
 	struct session	*s, *sloop, *stmp;
+	struct winlink	*wl;
 
-	if ((s = cmd_find_session(cmdq, args_get(args, 't'), 0)) == NULL)
-		return (CMD_RETURN_ERROR);
+	s = item->target.s;
 
-	if (args_has(args, 'a')) {
+	if (args_has(args, 'C')) {
+		RB_FOREACH(wl, winlinks, &s->windows) {
+			wl->window->flags &= ~WINDOW_ALERTFLAGS;
+			wl->flags &= ~WINLINK_ALERTFLAGS;
+		}
+		server_redraw_session(s);
+	} else if (args_has(args, 'a')) {
 		RB_FOREACH_SAFE(sloop, sessions, &sessions, stmp) {
 			if (sloop != s) {
 				server_destroy_session(sloop);
