@@ -1,4 +1,3 @@
-/* $Id$ */
 /*	$OpenBSD: daemon.c,v 1.6 2005/08/08 08:05:33 espie Exp $ */
 /*-
  * Copyright (c) 1990, 1993
@@ -33,7 +32,7 @@
 #include <unistd.h>
 #include <stdlib.h>
 
-#include "tmux.h"
+#include "compat.h"
 
 int
 daemon(int nochdir, int noclose)
