@@ -40,42 +40,54 @@
 struct format_entry;
 typedef void (*format_cb)(struct format_tree *, struct format_entry *);
 
-void	 format_job_callback(struct job *);
-char	*format_job_get(struct format_tree *, const char *);
-void	 format_job_timer(int, short, void *);
+static char	*format_job_get(struct format_tree *, const char *);
+static void	 format_job_timer(int, short, void *);
 
-void	 format_cb_host(struct format_tree *, struct format_entry *);
-void	 format_cb_host_short(struct format_tree *, struct format_entry *);
-void	 format_cb_pid(struct format_tree *, struct format_entry *);
-void	 format_cb_session_alerts(struct format_tree *, struct format_entry *);
-void	 format_cb_window_layout(struct format_tree *, struct format_entry *);
-void	 format_cb_window_visible_layout(struct format_tree *,
-	     struct format_entry *);
-void	 format_cb_start_command(struct format_tree *, struct format_entry *);
-void	 format_cb_current_command(struct format_tree *, struct format_entry *);
-void	 format_cb_current_path(struct format_tree *, struct format_entry *);
-void	 format_cb_history_bytes(struct format_tree *, struct format_entry *);
-void	 format_cb_pane_tabs(struct format_tree *, struct format_entry *);
+static void	 format_cb_host(struct format_tree *, struct format_entry *);
+static void	 format_cb_host_short(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_pid(struct format_tree *, struct format_entry *);
+static void	 format_cb_session_alerts(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_window_layout(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_window_visible_layout(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_start_command(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_current_command(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_history_bytes(struct format_tree *,
+		     struct format_entry *);
+static void	 format_cb_pane_tabs(struct format_tree *,
+		     struct format_entry *);
 
-char	*format_find(struct format_tree *, const char *, int);
-void	 format_add_cb(struct format_tree *, const char *, format_cb);
-void	 format_add_tv(struct format_tree *, const char *, struct timeval *);
-int	 format_replace(struct format_tree *, const char *, size_t, char **,
-	     size_t *, size_t *);
-char	*format_time_string(time_t);
+static void	 format_cb_current_path(struct format_tree *,
+		     struct format_entry *);
 
-void	 format_defaults_pane_tabs(struct format_tree *, struct window_pane *);
-void	 format_defaults_session(struct format_tree *, struct session *);
-void	 format_defaults_client(struct format_tree *, struct client *);
-void	 format_defaults_winlink(struct format_tree *, struct session *,
-	     struct winlink *);
+static char	*format_find(struct format_tree *, const char *, int);
+static void	 format_add_cb(struct format_tree *, const char *, format_cb);
+static void	 format_add_tv(struct format_tree *, const char *,
+		     struct timeval *);
+static int	 format_replace(struct format_tree *, const char *, size_t,
+		     char **, size_t *, size_t *);
+
+static void	 format_defaults_session(struct format_tree *,
+		     struct session *);
+static void	 format_defaults_client(struct format_tree *, struct client *);
+static void	 format_defaults_winlink(struct format_tree *,
+		     struct winlink *);
 
 /* Entry in format job tree. */
 struct format_job {
+	struct client		*client;
+	u_int			 tag;
 	const char		*cmd;
+	const char		*expanded;
 
 	time_t			 last;
 	char			*out;
+	int			 updated;
 
 	struct job		*job;
 	int			 status;
@@ -84,16 +96,19 @@ struct format_job {
 };
 
 /* Format job tree. */
-struct event format_job_event;
-int	format_job_cmp(struct format_job *, struct format_job *);
-RB_HEAD(format_job_tree, format_job) format_jobs = RB_INITIALIZER();
-RB_PROTOTYPE(format_job_tree, format_job, entry, format_job_cmp);
-RB_GENERATE(format_job_tree, format_job, entry, format_job_cmp);
+static struct event format_job_event;
+static int format_job_cmp(struct format_job *, struct format_job *);
+static RB_HEAD(format_job_tree, format_job) format_jobs = RB_INITIALIZER();
+RB_GENERATE_STATIC(format_job_tree, format_job, entry, format_job_cmp);
 
 /* Format job tree comparison function. */
-int
+static int
 format_job_cmp(struct format_job *fj1, struct format_job *fj2)
 {
+	if (fj1->tag < fj2->tag)
+		return (-1);
+	if (fj1->tag > fj2->tag)
+		return (1);
 	return (strcmp(fj1->cmd, fj2->cmd));
 }
 
@@ -115,26 +130,28 @@ struct format_entry {
 /* Format entry tree. */
 struct format_tree {
 	struct window		*w;
+	struct winlink		*wl;
 	struct session		*s;
 	struct window_pane	*wp;
 
+	struct client		*client;
+	u_int			 tag;
 	int			 flags;
 
 	RB_HEAD(format_entry_tree, format_entry) tree;
 };
-int	format_entry_cmp(struct format_entry *, struct format_entry *);
-RB_PROTOTYPE(format_entry_tree, format_entry, entry, format_entry_cmp);
-RB_GENERATE(format_entry_tree, format_entry, entry, format_entry_cmp);
+static int format_entry_cmp(struct format_entry *, struct format_entry *);
+RB_GENERATE_STATIC(format_entry_tree, format_entry, entry, format_entry_cmp);
 
 /* Format entry tree comparison function. */
-int
+static int
 format_entry_cmp(struct format_entry *fe1, struct format_entry *fe2)
 {
 	return (strcmp(fe1->key, fe2->key));
 }
 
 /* Single-character uppercase aliases. */
-const char *format_upper[] = {
+static const char *format_upper[] = {
 	NULL,		/* A */
 	NULL,		/* B */
 	NULL,		/* C */
@@ -164,7 +181,7 @@ const char *format_upper[] = {
 };
 
 /* Single-character lowercase aliases. */
-const char *format_lower[] = {
+static const char *format_lower[] = {
 	NULL,		/* a */
 	NULL,		/* b */
 	NULL,		/* c */
@@ -193,17 +210,41 @@ const char *format_lower[] = {
 	NULL		/* z */
 };
 
-/* Format job callback. */
-void
-format_job_callback(struct job *job)
+/* Format job update callback. */
+static void
+format_job_update(struct job *job)
+{
+	struct format_job	*fj = job->data;
+	char			*line;
+	time_t			 t;
+	struct client		*c;
+
+	if ((line = evbuffer_readline(job->event->input)) == NULL)
+		return;
+	fj->updated = 1;
+
+	free(fj->out);
+	fj->out = line;
+
+	log_debug("%s: %s: %s", __func__, fj->cmd, fj->out);
+
+	t = time (NULL);
+	if (fj->status && fj->last != t) {
+		TAILQ_FOREACH(c, &clients, entry)
+		    server_status_client(c);
+		fj->last = t;
+	}
+}
+
+/* Format job complete callback. */
+static void
+format_job_complete(struct job *job)
 {
 	struct format_job	*fj = job->data;
 	char			*line, *buf;
 	size_t			 len;
-	struct client		*c;
 
 	fj->job = NULL;
-	free(fj->out);
 
 	buf = NULL;
 	if ((line = evbuffer_readline(job->event->input)) == NULL) {
@@ -214,82 +255,135 @@ format_job_callback(struct job *job)
 		buf[len] = '\0';
 	} else
 		buf = line;
-	fj->out = buf;
+
+	if (*buf != '\0' || !fj->updated) {
+		free(fj->out);
+		fj->out = buf;
+		log_debug("%s: %s: %s", __func__, fj->cmd, fj->out);
+	} else
+		free(buf);
 
 	if (fj->status) {
-		TAILQ_FOREACH(c, &clients, entry)
-		    server_status_client(c);
+		if (fj->client != NULL)
+			server_status_client(fj->client);
 		fj->status = 0;
 	}
-
-	log_debug("%s: %s: %s", __func__, fj->cmd, fj->out);
 }
 
 /* Find a job. */
-char *
+static char *
 format_job_get(struct format_tree *ft, const char *cmd)
 {
-	struct format_job	fj0, *fj;
-	time_t			t;
+	struct format_job_tree	*jobs;
+	struct format_job	 fj0, *fj;
+	time_t			 t;
+	char			*expanded;
+	int			 force;
 
+	if (ft->client == NULL)
+		jobs = &format_jobs;
+	else if (ft->client->jobs != NULL)
+		jobs = ft->client->jobs;
+	else {
+		jobs = ft->client->jobs = xmalloc(sizeof *ft->client->jobs);
+		RB_INIT(jobs);
+	}
+
+	fj0.tag = ft->tag;
 	fj0.cmd = cmd;
-	if ((fj = RB_FIND(format_job_tree, &format_jobs, &fj0)) == NULL) {
+	if ((fj = RB_FIND(format_job_tree, jobs, &fj0)) == NULL) {
 		fj = xcalloc(1, sizeof *fj);
+		fj->client = ft->client;
+		fj->tag = ft->tag;
 		fj->cmd = xstrdup(cmd);
+		fj->expanded = NULL;
 
 		xasprintf(&fj->out, "<'%s' not ready>", fj->cmd);
 
-		RB_INSERT(format_job_tree, &format_jobs, fj);
+		RB_INSERT(format_job_tree, jobs, fj);
 	}
 
+	expanded = format_expand(ft, cmd);
+	if (fj->expanded == NULL || strcmp(expanded, fj->expanded) != 0) {
+		free((void *)fj->expanded);
+		fj->expanded = xstrdup(expanded);
+		force = 1;
+	} else
+		force = (ft->flags & FORMAT_FORCE);
+
 	t = time(NULL);
-	if (fj->job == NULL && ((ft->flags & FORMAT_FORCE) || fj->last != t)) {
-		fj->job = job_run(fj->cmd, NULL, NULL, format_job_callback,
-		    NULL, fj);
+	if (fj->job == NULL && (force || fj->last != t)) {
+		fj->job = job_run(expanded, NULL, NULL, format_job_update,
+		    format_job_complete, NULL, fj);
 		if (fj->job == NULL) {
 			free(fj->out);
 			xasprintf(&fj->out, "<'%s' didn't start>", fj->cmd);
 		}
 		fj->last = t;
+		fj->updated = 0;
 	}
 
 	if (ft->flags & FORMAT_STATUS)
 		fj->status = 1;
 
+	free(expanded);
 	return (format_expand(ft, fj->out));
 }
 
 /* Remove old jobs. */
-void
-format_job_timer(__unused int fd, __unused short events, __unused void *arg)
+static void
+format_job_tidy(struct format_job_tree *jobs, int force)
 {
 	struct format_job	*fj, *fj1;
 	time_t			 now;
-	struct timeval		 tv = { .tv_sec = 60 };
 
 	now = time(NULL);
-	RB_FOREACH_SAFE(fj, format_job_tree, &format_jobs, fj1) {
-		if (fj->last > now || now - fj->last < 3600)
+	RB_FOREACH_SAFE(fj, format_job_tree, jobs, fj1) {
+		if (!force && (fj->last > now || now - fj->last < 3600))
 			continue;
-		RB_REMOVE(format_job_tree, &format_jobs, fj);
+		RB_REMOVE(format_job_tree, jobs, fj);
 
 		log_debug("%s: %s", __func__, fj->cmd);
 
 		if (fj->job != NULL)
 			job_free(fj->job);
 
+		free((void *)fj->expanded);
 		free((void *)fj->cmd);
 		free(fj->out);
 
 		free(fj);
 	}
+}
+
+/* Remove old jobs for client. */
+void
+format_lost_client(struct client *c)
+{
+	if (c->jobs != NULL)
+		format_job_tidy(c->jobs, 1);
+	free(c->jobs);
+}
+
+/* Remove old jobs periodically. */
+static void
+format_job_timer(__unused int fd, __unused short events, __unused void *arg)
+{
+	struct client	*c;
+	struct timeval	 tv = { .tv_sec = 60 };
+
+	format_job_tidy(&format_jobs, 0);
+	TAILQ_FOREACH(c, &clients, entry) {
+		if (c->jobs != NULL)
+			format_job_tidy(c->jobs, 0);
+	}
 
 	evtimer_del(&format_job_event);
 	evtimer_add(&format_job_event, &tv);
 }
 
 /* Callback for host. */
-void
+static void
 format_cb_host(__unused struct format_tree *ft, struct format_entry *fe)
 {
 	char host[HOST_NAME_MAX + 1];
@@ -301,7 +395,7 @@ format_cb_host(__unused struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for host_short. */
-void
+static void
 format_cb_host_short(__unused struct format_tree *ft, struct format_entry *fe)
 {
 	char host[HOST_NAME_MAX + 1], *cp;
@@ -316,19 +410,19 @@ format_cb_host_short(__unused struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for pid. */
-void
+static void
 format_cb_pid(__unused struct format_tree *ft, struct format_entry *fe)
 {
 	xasprintf(&fe->value, "%ld", (long)getpid());
 }
 
 /* Callback for session_alerts. */
-void
+static void
 format_cb_session_alerts(struct format_tree *ft, struct format_entry *fe)
 {
 	struct session	*s = ft->s;
 	struct winlink	*wl;
-	char		 alerts[256], tmp[16];
+	char		 alerts[1024], tmp[16];
 
 	if (s == NULL)
 		return;
@@ -352,8 +446,50 @@ format_cb_session_alerts(struct format_tree *ft, struct format_entry *fe)
 	fe->value = xstrdup(alerts);
 }
 
+/* Callback for session_stack. */
+static void
+format_cb_session_stack(struct format_tree *ft, struct format_entry *fe)
+{
+	struct session	*s = ft->s;
+	struct winlink	*wl;
+	char		 result[1024], tmp[16];
+
+	if (s == NULL)
+		return;
+
+	xsnprintf(result, sizeof result, "%u", s->curw->idx);
+	TAILQ_FOREACH(wl, &s->lastw, sentry) {
+		xsnprintf(tmp, sizeof tmp, "%u", wl->idx);
+
+		if (*result != '\0')
+			strlcat(result, ",", sizeof result);
+		strlcat(result, tmp, sizeof result);
+	}
+	fe->value = xstrdup(result);
+}
+
+/* Callback for window_stack_index. */
+static void
+format_cb_window_stack_index(struct format_tree *ft, struct format_entry *fe)
+{
+	struct session	*s = ft->wl->session;
+	struct winlink	*wl;
+	u_int		 idx;
+
+	idx = 0;
+	TAILQ_FOREACH(wl, &s->lastw, sentry) {
+		idx++;
+		if (wl == ft->wl)
+			break;
+	}
+	if (wl != NULL)
+		xasprintf(&fe->value, "%u", idx);
+	else
+		fe->value = xstrdup("0");
+}
+
 /* Callback for window_layout. */
-void
+static void
 format_cb_window_layout(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window	*w = ft->w;
@@ -368,7 +504,7 @@ format_cb_window_layout(struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for window_visible_layout. */
-void
+static void
 format_cb_window_visible_layout(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window	*w = ft->w;
@@ -380,7 +516,7 @@ format_cb_window_visible_layout(struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for pane_start_command. */
-void
+static void
 format_cb_start_command(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window_pane	*wp = ft->wp;
@@ -392,7 +528,7 @@ format_cb_start_command(struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for pane_current_command. */
-void
+static void
 format_cb_current_command(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window_pane	*wp = ft->wp;
@@ -430,7 +566,7 @@ format_cb_current_path(struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for history_bytes. */
-void
+static void
 format_cb_history_bytes(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window_pane	*wp = ft->wp;
@@ -455,7 +591,7 @@ format_cb_history_bytes(struct format_tree *ft, struct format_entry *fe)
 }
 
 /* Callback for pane_tabs. */
-void
+static void
 format_cb_pane_tabs(struct format_tree *ft, struct format_entry *fe)
 {
 	struct window_pane	*wp = ft->wp;
@@ -480,12 +616,23 @@ format_cb_pane_tabs(struct format_tree *ft, struct format_entry *fe)
 	evbuffer_free(buffer);
 }
 
+/* Merge a format tree. */
+static void
+format_merge(struct format_tree *ft, struct format_tree *from)
+{
+	struct format_entry	*fe;
+
+	RB_FOREACH(fe, format_entry_tree, &from->tree) {
+		if (fe->value != NULL)
+			format_add(ft, fe->key, "%s", fe->value);
+	}
+}
+
 /* Create a new tree. */
 struct format_tree *
-format_create(struct cmd_q *cmdq, int flags)
+format_create(struct client *c, struct cmdq_item *item, int tag, int flags)
 {
 	struct format_tree	*ft;
-	struct cmd		*cmd;
 
 	if (!event_initialized(&format_job_event)) {
 		evtimer_set(&format_job_event, format_job_timer, NULL);
@@ -494,19 +641,27 @@ format_create(struct cmd_q *cmdq, int flags)
 
 	ft = xcalloc(1, sizeof *ft);
 	RB_INIT(&ft->tree);
+
+	if (c != NULL) {
+		ft->client = c;
+		ft->client->references++;
+	}
+
+	ft->tag = tag;
 	ft->flags = flags;
 
+	format_add(ft, "version", "%s", VERSION);
 	format_add_cb(ft, "host", format_cb_host);
 	format_add_cb(ft, "host_short", format_cb_host_short);
 	format_add_cb(ft, "pid", format_cb_pid);
 	format_add(ft, "socket_path", "%s", socket_path);
 	format_add_tv(ft, "start_time", &start_time);
 
-	if (cmdq != NULL && cmdq->cmd != NULL)
-		format_add(ft, "command_name", "%s", cmdq->cmd->entry->name);
-	if (cmdq != NULL && cmdq->parent != NULL) {
-		cmd = cmdq->parent->cmd;
-		format_add(ft, "command_hooked", "%s", cmd->entry->name);
+	if (item != NULL) {
+		if (item->cmd != NULL)
+			format_add(ft, "command", "%s", item->cmd->entry->name);
+		if (item->shared != NULL && item->shared->formats != NULL)
+			format_merge(ft, item->shared->formats);
 	}
 
 	return (ft);
@@ -525,6 +680,8 @@ format_free(struct format_tree *ft)
 		free(fe);
 	}
 
+	if (ft->client != NULL)
+		server_client_unref(ft->client);
 	free(ft);
 }
 
@@ -556,7 +713,7 @@ format_add(struct format_tree *ft, const char *key, const char *fmt, ...)
 }
 
 /* Add a key and time. */
-void
+static void
 format_add_tv(struct format_tree *ft, const char *key, struct timeval *tv)
 {
 	struct format_entry	*fe;
@@ -580,7 +737,7 @@ format_add_tv(struct format_tree *ft, const char *key, struct timeval *tv)
 }
 
 /* Add a key and function. */
-void
+static void
 format_add_cb(struct format_tree *ft, const char *key, format_cb cb)
 {
 	struct format_entry	*fe;
@@ -604,43 +761,33 @@ format_add_cb(struct format_tree *ft, const char *key, format_cb cb)
 }
 
 /* Find a format entry. */
-char *
+static char *
 format_find(struct format_tree *ft, const char *key, int modifiers)
 {
 	struct format_entry	*fe, fe_find;
-	struct options_entry	*o;
 	struct environ_entry	*envent;
 	static char		 s[64];
+	struct options_entry	*o;
 	const char		*found;
+	int			 idx;
 	char			*copy, *saved;
 
-	found = NULL;
-
 	if (~modifiers & FORMAT_TIMESTRING) {
-		o = options_find(global_options, key);
+		o = options_parse_get(global_options, key, &idx, 0);
 		if (o == NULL && ft->w != NULL)
-			o = options_find(ft->w->options, key);
+			o = options_parse_get(ft->w->options, key, &idx, 0);
 		if (o == NULL)
-			o = options_find(global_w_options, key);
+			o = options_parse_get(global_w_options, key, &idx, 0);
 		if (o == NULL && ft->s != NULL)
-			o = options_find(ft->s->options, key);
+			o = options_parse_get(ft->s->options, key, &idx, 0);
 		if (o == NULL)
-			o = options_find(global_s_options, key);
+			o = options_parse_get(global_s_options, key, &idx, 0);
 		if (o != NULL) {
-			switch (o->type) {
-			case OPTIONS_STRING:
-				found = o->str;
-				goto found;
-			case OPTIONS_NUMBER:
-				xsnprintf(s, sizeof s, "%lld", o->num);
-				found = s;
-				goto found;
-			case OPTIONS_STYLE:
-				found = style_tostring(&o->style);
-				goto found;
-			}
+			found = options_tostring(o, idx, 1);
+			goto found;
 		}
 	}
+	found = NULL;
 
 	fe_find.key = (char *) key;
 	fe = RB_FIND(format_entry_tree, &ft->tree, &fe_find);
@@ -695,19 +842,63 @@ found:
 	return (copy);
 }
 
+/* Skip until comma. */
+static char *
+format_skip(char *s)
+{
+	int	brackets = 0;
+
+	for (; *s != '\0'; s++) {
+		if (*s == '{')
+			brackets++;
+		if (*s == '}')
+			brackets--;
+		if (*s == ',' && brackets == 0)
+			break;
+	}
+	if (*s == '\0')
+		return (NULL);
+	return (s);
+}
+
+/* Return left and right alternatives separated by commas. */
+static int
+format_choose(char *s, char **left, char **right)
+{
+	char	*cp;
+
+	cp = format_skip(s);
+	if (cp == NULL)
+		return (-1);
+	*cp = '\0';
+
+	*left = s;
+	*right = cp + 1;
+	return (0);
+}
+
+/* Is this true? */
+static int
+format_true(const char *s)
+{
+	if (s != NULL && *s != '\0' && (s[0] != '0' || s[1] != '\0'))
+		return (1);
+	return (0);
+}
+
 /*
  * Replace a key/value pair in buffer. #{blah} is expanded directly,
  * #{?blah,a,b} is replace with a if blah exists and is nonzero else b.
  */
-int
+static int
 format_replace(struct format_tree *ft, const char *key, size_t keylen,
     char **buf, size_t *len, size_t *off)
 {
 	char		*copy, *copy0, *endptr, *ptr, *found, *new, *value;
-	char		*from = NULL, *to = NULL;
+	char		*from = NULL, *to = NULL, *left, *right;
 	size_t		 valuelen, newlen, fromlen, tolen, used;
 	long		 limit = 0;
-	int		 modifiers = 0, brackets;
+	int		 modifiers = 0, compare = 0;
 
 	/* Make a copy of the key. */
 	copy0 = copy = xmalloc(keylen + 1);
@@ -716,7 +907,19 @@ format_replace(struct format_tree *ft, const char *key, size_t keylen,
 
 	/* Is there a length limit or whatnot? */
 	switch (copy[0]) {
+	case '!':
+		if (copy[1] == '=' && copy[2] == ':') {
+			compare = -1;
+			copy += 3;
+			break;
+		}
+		break;
 	case '=':
+		if (copy[1] == '=' && copy[2] == ':') {
+			compare = 1;
+			copy += 3;
+			break;
+		}
 		errno = 0;
 		limit = strtol(copy + 1, &endptr, 10);
 		if (errno == ERANGE && (limit == LONG_MIN || limit == LONG_MAX))
@@ -768,39 +971,41 @@ format_replace(struct format_tree *ft, const char *key, size_t keylen,
 		break;
 	}
 
-	/*
-	 * Is this a conditional? If so, check it exists and extract either the
-	 * first or second element. If not, look up the key directly.
-	 */
-	if (*copy == '?') {
-		ptr = strchr(copy, ',');
+	/* Is this a comparison or a conditional? */
+	if (compare != 0) {
+		/* Comparison: compare comma-separated left and right. */
+		if (format_choose(copy, &left, &right) != 0)
+			goto fail;
+		left = format_expand(ft, left);
+		right = format_expand(ft, right);
+		if (compare == 1 && strcmp(left, right) == 0)
+			value = xstrdup("1");
+		else if (compare == -1 && strcmp(left, right) != 0)
+			value = xstrdup("1");
+		else
+			value = xstrdup("0");
+		free(right);
+		free(left);
+	} else if (*copy == '?') {
+		/* Conditional: check first and choose second or third. */
+		ptr = format_skip(copy);
 		if (ptr == NULL)
 			goto fail;
 		*ptr = '\0';
 
-		value = ptr + 1;
 		found = format_find(ft, copy + 1, modifiers);
-
-		brackets = 0;
-		for (ptr = ptr + 1; *ptr != '\0'; ptr++) {
-			if (*ptr == '{')
-				brackets++;
-			if (*ptr == '}')
-				brackets--;
-			if (*ptr == ',' && brackets == 0)
-				break;
-		}
-		if (*ptr == '\0')
+		if (found == NULL)
+			found = format_expand(ft, copy + 1);
+		if (format_choose(ptr + 1, &left, &right) != 0)
 			goto fail;
 
-		if (found != NULL && *found != '\0' &&
-		    (found[0] != '0' || found[1] != '\0')) {
-			*ptr = '\0';
-		} else
-			value = ptr + 1;
-		value = format_expand(ft, value);
+		if (format_true(found))
+			value = format_expand(ft, left);
+		else
+			value = format_expand(ft, right);
 		free(found);
 	} else {
+		/* Neither: look up directly. */
 		value = format_find(ft, copy, modifiers);
 		if (value == NULL)
 			value = xstrdup("");
@@ -885,7 +1090,7 @@ format_expand_time(struct format_tree *ft, const char *fmt, time_t t)
 char *
 format_expand(struct format_tree *ft, const char *fmt)
 {
-	char		*buf, *tmp, *cmd, *out;
+	char		*buf, *out;
 	const char	*ptr, *s, *saved = fmt;
 	size_t		 off, len, n, outlen;
 	int     	 ch, brackets;
@@ -922,17 +1127,12 @@ format_expand(struct format_tree *ft, const char *fmt)
 				break;
 			n = ptr - fmt;
 
-			tmp = xmalloc(n + 1);
-			memcpy(tmp, fmt, n);
-			tmp[n] = '\0';
-			cmd = format_expand(ft, tmp);
-
-			out = format_job_get(ft, cmd);
+			if (ft->flags & FORMAT_NOJOBS)
+				out = xstrdup("");
+			else
+				out = format_job_get(ft, xstrndup(fmt, n));
 			outlen = strlen(out);
 
-			free(cmd);
-			free(tmp);
-
 			while (len - off < outlen + 1) {
 				buf = xreallocarray(buf, 2, len);
 				len *= 2;
@@ -996,6 +1196,25 @@ format_expand(struct format_tree *ft, const char *fmt)
 	return (buf);
 }
 
+/* Expand a single string. */
+char *
+format_single(struct cmdq_item *item, const char *fmt, struct client *c,
+    struct session *s, struct winlink *wl, struct window_pane *wp)
+{
+	struct format_tree	*ft;
+	char			*expanded;
+
+	if (item != NULL)
+		ft = format_create(item->client, item, FORMAT_NONE, 0);
+	else
+		ft = format_create(NULL, item, FORMAT_NONE, 0);
+	format_defaults(ft, c, s, wl, wp);
+
+	expanded = format_expand(ft, fmt);
+	format_free(ft);
+	return (expanded);
+}
+
 /* Set defaults for any of arguments that are not NULL. */
 void
 format_defaults(struct format_tree *ft, struct client *c, struct session *s,
@@ -1012,14 +1231,14 @@ format_defaults(struct format_tree *ft, struct client *c, struct session *s,
 		format_defaults_client(ft, c);
 	if (s != NULL)
 		format_defaults_session(ft, s);
-	if (s != NULL && wl != NULL)
-		format_defaults_winlink(ft, s, wl);
+	if (wl != NULL)
+		format_defaults_winlink(ft, wl);
 	if (wp != NULL)
 		format_defaults_pane(ft, wp);
 }
 
 /* Set default format keys for a session. */
-void
+static void
 format_defaults_session(struct format_tree *ft, struct session *s)
 {
 	struct session_group	*sg;
@@ -1032,10 +1251,10 @@ format_defaults_session(struct format_tree *ft, struct session *s)
 	format_add(ft, "session_height", "%u", s->sy);
 	format_add(ft, "session_id", "$%u", s->id);
 
-	sg = session_group_find(s);
+	sg = session_group_contains(s);
 	format_add(ft, "session_grouped", "%d", sg != NULL);
 	if (sg != NULL)
-		format_add(ft, "session_group", "%u", session_group_index(sg));
+		format_add(ft, "session_group", "%s", sg->name);
 
 	format_add_tv(ft, "session_created", &s->creation_time);
 	format_add_tv(ft, "session_last_attached", &s->last_attached_time);
@@ -1045,31 +1264,40 @@ format_defaults_session(struct format_tree *ft, struct session *s)
 	format_add(ft, "session_many_attached", "%d", s->attached > 1);
 
 	format_add_cb(ft, "session_alerts", format_cb_session_alerts);
+	format_add_cb(ft, "session_stack", format_cb_session_stack);
 }
 
 /* Set default format keys for a client. */
-void
+static void
 format_defaults_client(struct format_tree *ft, struct client *c)
 {
 	struct session	*s;
 	const char	*name;
+	struct tty	*tty = &c->tty;
+	const char	*types[] = TTY_TYPES;
 
 	if (ft->s == NULL)
 		ft->s = c->session;
 
+	format_add(ft, "client_name", "%s", c->name);
 	format_add(ft, "client_pid", "%ld", (long) c->pid);
-	format_add(ft, "client_height", "%u", c->tty.sy);
-	format_add(ft, "client_width", "%u", c->tty.sx);
-	if (c->tty.path != NULL)
-		format_add(ft, "client_tty", "%s", c->tty.path);
-	if (c->tty.termname != NULL)
-		format_add(ft, "client_termname", "%s", c->tty.termname);
+	format_add(ft, "client_height", "%u", tty->sy);
+	format_add(ft, "client_width", "%u", tty->sx);
+	format_add(ft, "client_tty", "%s", c->ttyname);
 	format_add(ft, "client_control_mode", "%d",
 		!!(c->flags & CLIENT_CONTROL));
 
+	if (tty->term_name != NULL)
+		format_add(ft, "client_termname", "%s", tty->term_name);
+	if (tty->term_name != NULL)
+		format_add(ft, "client_termtype", "%s", types[tty->term_type]);
+
 	format_add_tv(ft, "client_created", &c->creation_time);
 	format_add_tv(ft, "client_activity", &c->activity_time);
 
+	format_add(ft, "client_written", "%zu", c->written);
+	format_add(ft, "client_discarded", "%zu", c->discarded);
+
 	name = server_client_get_key_table(c);
 	if (strcmp(c->keytable->name, name) == 0)
 		format_add(ft, "client_prefix", "%d", 0);
@@ -1077,7 +1305,7 @@ format_defaults_client(struct format_tree *ft, struct client *c)
 		format_add(ft, "client_prefix", "%d", 1);
 	format_add(ft, "client_key_table", "%s", c->keytable->name);
 
-	if (c->tty.flags & TTY_UTF8)
+	if (tty->flags & TTY_UTF8)
 		format_add(ft, "client_utf8", "%d", 1);
 	else
 		format_add(ft, "client_utf8", "%d", 0);
@@ -1115,22 +1343,21 @@ format_defaults_window(struct format_tree *ft, struct window *w)
 }
 
 /* Set default format keys for a winlink. */
-void
-format_defaults_winlink(struct format_tree *ft, struct session *s,
-    struct winlink *wl)
+static void
+format_defaults_winlink(struct format_tree *ft, struct winlink *wl)
 {
+	struct session	*s = wl->session;
 	struct window	*w = wl->window;
-	char		*flags;
 
 	if (ft->w == NULL)
 		ft->w = wl->window;
-
-	flags = window_printable_flags(s, wl);
+	ft->wl = wl;
 
 	format_defaults_window(ft, w);
 
 	format_add(ft, "window_index", "%d", wl->idx);
-	format_add(ft, "window_flags", "%s", flags);
+	format_add_cb(ft, "window_stack_index", format_cb_window_stack_index);
+	format_add(ft, "window_flags", "%s", window_printable_flags(wl));
 	format_add(ft, "window_active", "%d", wl == s->curw);
 
 	format_add(ft, "window_bell_flag", "%d",
@@ -1142,8 +1369,6 @@ format_defaults_winlink(struct format_tree *ft, struct session *s,
 	format_add(ft, "window_last_flag", "%d",
 	    !!(wl == TAILQ_FIRST(&s->lastw)));
 	format_add(ft, "window_linked", "%d", session_is_linked(s, wl->window));
-
-	free(flags);
 }
 
 /* Set default format keys for a window pane. */
@@ -1186,8 +1411,13 @@ format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 	}
 
 	format_add(ft, "pane_in_mode", "%d", wp->screen != &wp->base);
+	if (wp->mode != NULL)
+		format_add(ft, "pane_mode", "%s", wp->mode->name);
+
 	format_add(ft, "pane_synchronized", "%d",
 	    !!options_get_number(wp->window->options, "synchronize-panes"));
+	format_add(ft, "pane_search_string", "%s",
+	    window_copy_search_string(wp));
 
 	format_add(ft, "pane_tty", "%s", wp->tty);
 	format_add(ft, "pane_pid", "%ld", (long) wp->pid);
@@ -1220,11 +1450,13 @@ format_defaults_pane(struct format_tree *ft, struct window_pane *wp)
 	    !!(wp->base.mode & MODE_WRAP));
 
 	format_add(ft, "mouse_any_flag", "%d",
-	    !!(wp->base.mode & (MODE_MOUSE_STANDARD|MODE_MOUSE_BUTTON)));
+	    !!(wp->base.mode & ALL_MOUSE_MODES));
 	format_add(ft, "mouse_standard_flag", "%d",
 	    !!(wp->base.mode & MODE_MOUSE_STANDARD));
 	format_add(ft, "mouse_button_flag", "%d",
 	    !!(wp->base.mode & MODE_MOUSE_BUTTON));
+	format_add(ft, "mouse_all_flag", "%d",
+	    !!(wp->base.mode & MODE_MOUSE_ALL));
 
 	format_add_cb(ft, "pane_tabs", format_cb_pane_tabs);
 }
