@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -30,10 +30,10 @@
  * direction with output).
  */
 
-void	 input_key_mouse(struct window_pane *, struct mouse_event *);
+static void	 input_key_mouse(struct window_pane *, struct mouse_event *);
 
 struct input_key_ent {
-	int		 key;
+	key_code	 key;
 	const char	*data;
 
 	int		 flags;
@@ -41,7 +41,7 @@ struct input_key_ent {
 #define INPUTKEY_CURSOR 0x2	/* cursor key */
 };
 
-const struct input_key_ent input_keys[] = {
+static const struct input_key_ent input_keys[] = {
 	/* Backspace key. */
 	{ KEYC_BSPACE,		"\177",		0 },
 
@@ -134,17 +134,32 @@ const struct input_key_ent input_keys[] = {
 	{ KEYC_KP_PERIOD,	".",		0 },
 };
 
+/* Split a character into two UTF-8 bytes. */
+static size_t
+input_split2(u_int c, u_char *dst)
+{
+	if (c > 0x7f) {
+		dst[0] = (c >> 6) | 0xc0;
+		dst[1] = (c & 0x3f) | 0x80;
+		return (2);
+	}
+	dst[0] = c;
+	return (1);
+}
+
 /* Translate a key code into an output key sequence. */
 void
-input_key(struct window_pane *wp, int key, struct mouse_event *m)
+input_key(struct window_pane *wp, key_code key, struct mouse_event *m)
 {
-	const struct input_key_ent     *ike;
-	u_int				i;
-	size_t				dlen;
-	char			       *out;
-	u_char				ch;
+	const struct input_key_ent	*ike;
+	u_int				 i;
+	size_t				 dlen;
+	char				*out;
+	key_code			 justkey;
+	struct utf8_data		 ud;
 
-	log_debug("writing key 0x%x (%s)", key, key_string_lookup_key(key));
+	log_debug("writing key 0x%llx (%s) to %%%u", key,
+	    key_string_lookup_key(key), wp->id);
 
 	/* If this is a mouse key, pass off to mouse function. */
 	if (KEYC_IS_MOUSE(key)) {
@@ -155,13 +170,22 @@ input_key(struct window_pane *wp, int key, struct mouse_event *m)
 
 	/*
 	 * If this is a normal 7-bit key, just send it, with a leading escape
-	 * if necessary.
+	 * if necessary. If it is a UTF-8 key, split it and send it.
 	 */
-	if (key != KEYC_NONE && (key & ~KEYC_ESCAPE) < 0x100) {
+	justkey = (key & ~KEYC_ESCAPE);
+	if (justkey <= 0x7f) {
 		if (key & KEYC_ESCAPE)
 			bufferevent_write(wp->event, "\033", 1);
-		ch = key & ~KEYC_ESCAPE;
-		bufferevent_write(wp->event, &ch, 1);
+		ud.data[0] = justkey;
+		bufferevent_write(wp->event, &ud.data[0], 1);
+		return;
+	}
+	if (justkey > 0x7f && justkey < KEYC_BASE) {
+		if (utf8_split(justkey, &ud) != UTF8_DONE)
+			return;
+		if (key & KEYC_ESCAPE)
+			bufferevent_write(wp->event, "\033", 1);
+		bufferevent_write(wp->event, ud.data, ud.size);
 		return;
 	}
 
@@ -169,13 +193,14 @@ input_key(struct window_pane *wp, int key, struct mouse_event *m)
 	 * Then try to look this up as an xterm key, if the flag to output them
 	 * is set.
 	 */
-	if (options_get_number(&wp->window->options, "xterm-keys")) {
+	if (options_get_number(wp->window->options, "xterm-keys")) {
 		if ((out = xterm_keys_lookup(key)) != NULL) {
 			bufferevent_write(wp->event, out, strlen(out));
 			free(out);
 			return;
 		}
 	}
+	key &= ~KEYC_XTERM;
 
 	/* Otherwise look the key up in the table. */
 	for (i = 0; i < nitems(input_keys); i++) {
@@ -194,11 +219,11 @@ input_key(struct window_pane *wp, int key, struct mouse_event *m)
 			break;
 	}
 	if (i == nitems(input_keys)) {
-		log_debug("key 0x%x missing", key);
+		log_debug("key 0x%llx missing", key);
 		return;
 	}
 	dlen = strlen(ike->data);
-	log_debug("found key 0x%x: \"%s\"", key, ike->data);
+	log_debug("found key 0x%llx: \"%s\"", key, ike->data);
 
 	/* Prefix a \033 for escape. */
 	if (key & KEYC_ESCAPE)
@@ -207,23 +232,45 @@ input_key(struct window_pane *wp, int key, struct mouse_event *m)
 }
 
 /* Translate mouse and output. */
-void
+static void
 input_key_mouse(struct window_pane *wp, struct mouse_event *m)
 {
-	char	buf[40];
-	size_t	len;
-	u_int	x, y;
+	struct screen	*s = wp->screen;
+	int		 mode = s->mode;
+	char		 buf[40];
+	size_t		 len;
+	u_int		 x, y;
 
-	if ((wp->screen->mode & ALL_MOUSE_MODES) == 0)
+	if ((mode & ALL_MOUSE_MODES) == 0)
 		return;
 	if (!window_pane_visible(wp))
 		return;
 	if (cmd_mouse_at(wp, m, &x, &y, 0) != 0)
 		return;
 
-	/* If this pane is not in button mode, discard motion events. */
-	if (!(wp->screen->mode & MODE_MOUSE_BUTTON) && (m->b & MOUSE_MASK_DRAG))
-		return;
+	/* If this pane is not in button or all mode, discard motion events. */
+	if (MOUSE_DRAG(m->b) &&
+	    (mode & (MODE_MOUSE_BUTTON|MODE_MOUSE_ALL)) == 0)
+	    return;
+
+	/*
+	 * If this event is a release event and not in all mode, discard it.
+	 * In SGR mode we can tell absolutely because a release is normally
+	 * shown by the last character. Without SGR, we check if the last
+	 * buttons was also a release.
+	 */
+	if (m->sgr_type != ' ') {
+		if (MOUSE_DRAG(m->sgr_b) &&
+		    MOUSE_BUTTONS(m->sgr_b) == 3 &&
+		    (~mode & MODE_MOUSE_ALL))
+			return;
+	} else {
+		if (MOUSE_DRAG(m->b) &&
+		    MOUSE_BUTTONS(m->b) == 3 &&
+		    MOUSE_BUTTONS(m->lb) == 3 &&
+		    (~mode & MODE_MOUSE_ALL))
+			return;
+	}
 
 	/*
 	 * Use the SGR (1006) extension only if the application requested it
@@ -234,14 +281,16 @@ input_key_mouse(struct window_pane *wp, struct mouse_event *m)
 	 * UTF-8 (1005) extension if the application requested, or to the
 	 * legacy format.
 	 */
-	if (m->sgr_type != ' ' && (wp->screen->mode & MODE_MOUSE_SGR)) {
+	if (m->sgr_type != ' ' && (s->mode & MODE_MOUSE_SGR)) {
 		len = xsnprintf(buf, sizeof buf, "\033[<%u;%u;%u%c",
 		    m->sgr_b, x + 1, y + 1, m->sgr_type);
-	} else if (wp->screen->mode & MODE_MOUSE_UTF8) {
+	} else if (s->mode & MODE_MOUSE_UTF8) {
+		if (m->b > 0x7ff - 32 || x > 0x7ff - 33 || y > 0x7ff - 33)
+			return;
 		len = xsnprintf(buf, sizeof buf, "\033[M");
-		len += utf8_split2(m->b + 32, &buf[len]);
-		len += utf8_split2(x + 33, &buf[len]);
-		len += utf8_split2(y + 33, &buf[len]);
+		len += input_split2(m->b + 32, &buf[len]);
+		len += input_split2(x + 33, &buf[len]);
+		len += input_split2(y + 33, &buf[len]);
 	} else {
 		if (m->b > 223)
 			return;
@@ -250,6 +299,6 @@ input_key_mouse(struct window_pane *wp, struct mouse_event *m)
 		buf[len++] = x + 33;
 		buf[len++] = y + 33;
 	}
-	log_debug("writing mouse %.*s", (int)len, buf);
+	log_debug("writing mouse %.*s to %%%u", (int)len, buf, wp->id);
 	bufferevent_write(wp->event, buf, len);
 }
