@@ -18,6 +18,9 @@
 
 #include <sys/types.h>
 
+#include <netinet/in.h>
+
+#include <resolv.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
@@ -84,6 +87,7 @@ struct input_ctx {
 	struct utf8_data	utf8data;
 
 	int			ch;
+
 	int			flags;
 #define INPUT_DISCARD 0x1
 
@@ -98,45 +102,50 @@ struct input_ctx {
 
 /* Helper functions. */
 struct input_transition;
-int	input_split(struct input_ctx *);
-int	input_get(struct input_ctx *, u_int, int, int);
-void printflike(2, 3) input_reply(struct input_ctx *, const char *, ...);
-void	input_set_state(struct window_pane *, const struct input_transition *);
-void	input_reset_cell(struct input_ctx *);
+static int	input_split(struct input_ctx *);
+static int	input_get(struct input_ctx *, u_int, int, int);
+static void printflike(2, 3) input_reply(struct input_ctx *, const char *, ...);
+static void	input_set_state(struct window_pane *,
+		    const struct input_transition *);
+static void	input_reset_cell(struct input_ctx *);
+
+static void	input_osc_4(struct window_pane *, const char *);
+static void	input_osc_52(struct window_pane *, const char *);
+static void	input_osc_104(struct window_pane *, const char *);
 
 /* Transition entry/exit handlers. */
-void	input_clear(struct input_ctx *);
-void	input_ground(struct input_ctx *);
-void	input_enter_osc(struct input_ctx *);
-void	input_exit_osc(struct input_ctx *);
-void	input_enter_apc(struct input_ctx *);
-void	input_exit_apc(struct input_ctx *);
-void	input_enter_rename(struct input_ctx *);
-void	input_exit_rename(struct input_ctx *);
+static void	input_clear(struct input_ctx *);
+static void	input_ground(struct input_ctx *);
+static void	input_enter_osc(struct input_ctx *);
+static void	input_exit_osc(struct input_ctx *);
+static void	input_enter_apc(struct input_ctx *);
+static void	input_exit_apc(struct input_ctx *);
+static void	input_enter_rename(struct input_ctx *);
+static void	input_exit_rename(struct input_ctx *);
 
 /* Input state handlers. */
-int	input_print(struct input_ctx *);
-int	input_intermediate(struct input_ctx *);
-int	input_parameter(struct input_ctx *);
-int	input_input(struct input_ctx *);
-int	input_c0_dispatch(struct input_ctx *);
-int	input_esc_dispatch(struct input_ctx *);
-int	input_csi_dispatch(struct input_ctx *);
-void	input_csi_dispatch_rm(struct input_ctx *);
-void	input_csi_dispatch_rm_private(struct input_ctx *);
-void	input_csi_dispatch_sm(struct input_ctx *);
-void	input_csi_dispatch_sm_private(struct input_ctx *);
-void	input_csi_dispatch_winops(struct input_ctx *);
-void	input_csi_dispatch_sgr_256(struct input_ctx *, int, u_int *);
-void	input_csi_dispatch_sgr_rgb(struct input_ctx *, int, u_int *);
-void	input_csi_dispatch_sgr(struct input_ctx *);
-int	input_dcs_dispatch(struct input_ctx *);
-int	input_utf8_open(struct input_ctx *);
-int	input_utf8_add(struct input_ctx *);
-int	input_utf8_close(struct input_ctx *);
+static int	input_print(struct input_ctx *);
+static int	input_intermediate(struct input_ctx *);
+static int	input_parameter(struct input_ctx *);
+static int	input_input(struct input_ctx *);
+static int	input_c0_dispatch(struct input_ctx *);
+static int	input_esc_dispatch(struct input_ctx *);
+static int	input_csi_dispatch(struct input_ctx *);
+static void	input_csi_dispatch_rm(struct input_ctx *);
+static void	input_csi_dispatch_rm_private(struct input_ctx *);
+static void	input_csi_dispatch_sm(struct input_ctx *);
+static void	input_csi_dispatch_sm_private(struct input_ctx *);
+static void	input_csi_dispatch_winops(struct input_ctx *);
+static void	input_csi_dispatch_sgr_256(struct input_ctx *, int, u_int *);
+static void	input_csi_dispatch_sgr_rgb(struct input_ctx *, int, u_int *);
+static void	input_csi_dispatch_sgr(struct input_ctx *);
+static int	input_dcs_dispatch(struct input_ctx *);
+static int	input_utf8_open(struct input_ctx *);
+static int	input_utf8_add(struct input_ctx *);
+static int	input_utf8_close(struct input_ctx *);
 
 /* Command table comparison function. */
-int	input_table_compare(const void *, const void *);
+static int	input_table_compare(const void *, const void *);
 
 /* Command table entry. */
 struct input_table_entry {
@@ -161,10 +170,11 @@ enum input_esc_type {
 	INPUT_ESC_SCSG0_ON,
 	INPUT_ESC_SCSG1_OFF,
 	INPUT_ESC_SCSG1_ON,
+	INPUT_ESC_ST,
 };
 
 /* Escape command table. */
-const struct input_table_entry input_esc_table[] = {
+static const struct input_table_entry input_esc_table[] = {
 	{ '0', "(", INPUT_ESC_SCSG0_ON },
 	{ '0', ")", INPUT_ESC_SCSG1_ON },
 	{ '7', "",  INPUT_ESC_DECSC },
@@ -178,6 +188,7 @@ const struct input_table_entry input_esc_table[] = {
 	{ 'E', "",  INPUT_ESC_NEL },
 	{ 'H', "",  INPUT_ESC_HTS },
 	{ 'M', "",  INPUT_ESC_RI },
+	{ '\\', "", INPUT_ESC_ST },
 	{ 'c', "",  INPUT_ESC_RIS },
 };
 
@@ -211,13 +222,14 @@ enum input_csi_type {
 	INPUT_CSI_SGR,
 	INPUT_CSI_SM,
 	INPUT_CSI_SM_PRIVATE,
+	INPUT_CSI_SU,
 	INPUT_CSI_TBC,
 	INPUT_CSI_VPA,
 	INPUT_CSI_WINOPS,
 };
 
 /* Control (CSI) command table. */
-const struct input_table_entry input_csi_table[] = {
+static const struct input_table_entry input_csi_table[] = {
 	{ '@', "",  INPUT_CSI_ICH },
 	{ 'A', "",  INPUT_CSI_CUU },
 	{ 'B', "",  INPUT_CSI_CUD },
@@ -232,6 +244,7 @@ const struct input_table_entry input_csi_table[] = {
 	{ 'L', "",  INPUT_CSI_IL },
 	{ 'M', "",  INPUT_CSI_DL },
 	{ 'P', "",  INPUT_CSI_DCH },
+	{ 'S', "",  INPUT_CSI_SU },
 	{ 'X', "",  INPUT_CSI_ECH },
 	{ 'Z', "",  INPUT_CSI_CBT },
 	{ 'c', "",  INPUT_CSI_DA },
@@ -276,169 +289,169 @@ struct input_state {
 	{ 0x1b, 0x1b, NULL,		 &input_state_esc_enter }
 
 /* Forward declarations of state tables. */
-const struct input_transition input_state_ground_table[];
-const struct input_transition input_state_esc_enter_table[];
-const struct input_transition input_state_esc_intermediate_table[];
-const struct input_transition input_state_csi_enter_table[];
-const struct input_transition input_state_csi_parameter_table[];
-const struct input_transition input_state_csi_intermediate_table[];
-const struct input_transition input_state_csi_ignore_table[];
-const struct input_transition input_state_dcs_enter_table[];
-const struct input_transition input_state_dcs_parameter_table[];
-const struct input_transition input_state_dcs_intermediate_table[];
-const struct input_transition input_state_dcs_handler_table[];
-const struct input_transition input_state_dcs_escape_table[];
-const struct input_transition input_state_dcs_ignore_table[];
-const struct input_transition input_state_osc_string_table[];
-const struct input_transition input_state_apc_string_table[];
-const struct input_transition input_state_rename_string_table[];
-const struct input_transition input_state_consume_st_table[];
-const struct input_transition input_state_utf8_three_table[];
-const struct input_transition input_state_utf8_two_table[];
-const struct input_transition input_state_utf8_one_table[];
+static const struct input_transition input_state_ground_table[];
+static const struct input_transition input_state_esc_enter_table[];
+static const struct input_transition input_state_esc_intermediate_table[];
+static const struct input_transition input_state_csi_enter_table[];
+static const struct input_transition input_state_csi_parameter_table[];
+static const struct input_transition input_state_csi_intermediate_table[];
+static const struct input_transition input_state_csi_ignore_table[];
+static const struct input_transition input_state_dcs_enter_table[];
+static const struct input_transition input_state_dcs_parameter_table[];
+static const struct input_transition input_state_dcs_intermediate_table[];
+static const struct input_transition input_state_dcs_handler_table[];
+static const struct input_transition input_state_dcs_escape_table[];
+static const struct input_transition input_state_dcs_ignore_table[];
+static const struct input_transition input_state_osc_string_table[];
+static const struct input_transition input_state_apc_string_table[];
+static const struct input_transition input_state_rename_string_table[];
+static const struct input_transition input_state_consume_st_table[];
+static const struct input_transition input_state_utf8_three_table[];
+static const struct input_transition input_state_utf8_two_table[];
+static const struct input_transition input_state_utf8_one_table[];
 
 /* ground state definition. */
-const struct input_state input_state_ground = {
+static const struct input_state input_state_ground = {
 	"ground",
 	input_ground, NULL,
 	input_state_ground_table
 };
 
 /* esc_enter state definition. */
-const struct input_state input_state_esc_enter = {
+static const struct input_state input_state_esc_enter = {
 	"esc_enter",
 	input_clear, NULL,
 	input_state_esc_enter_table
 };
 
 /* esc_intermediate state definition. */
-const struct input_state input_state_esc_intermediate = {
+static const struct input_state input_state_esc_intermediate = {
 	"esc_intermediate",
 	NULL, NULL,
 	input_state_esc_intermediate_table
 };
 
 /* csi_enter state definition. */
-const struct input_state input_state_csi_enter = {
+static const struct input_state input_state_csi_enter = {
 	"csi_enter",
 	input_clear, NULL,
 	input_state_csi_enter_table
 };
 
 /* csi_parameter state definition. */
-const struct input_state input_state_csi_parameter = {
+static const struct input_state input_state_csi_parameter = {
 	"csi_parameter",
 	NULL, NULL,
 	input_state_csi_parameter_table
 };
 
 /* csi_intermediate state definition. */
-const struct input_state input_state_csi_intermediate = {
+static const struct input_state input_state_csi_intermediate = {
 	"csi_intermediate",
 	NULL, NULL,
 	input_state_csi_intermediate_table
 };
 
 /* csi_ignore state definition. */
-const struct input_state input_state_csi_ignore = {
+static const struct input_state input_state_csi_ignore = {
 	"csi_ignore",
 	NULL, NULL,
 	input_state_csi_ignore_table
 };
 
 /* dcs_enter state definition. */
-const struct input_state input_state_dcs_enter = {
+static const struct input_state input_state_dcs_enter = {
 	"dcs_enter",
 	input_clear, NULL,
 	input_state_dcs_enter_table
 };
 
 /* dcs_parameter state definition. */
-const struct input_state input_state_dcs_parameter = {
+static const struct input_state input_state_dcs_parameter = {
 	"dcs_parameter",
 	NULL, NULL,
 	input_state_dcs_parameter_table
 };
 
 /* dcs_intermediate state definition. */
-const struct input_state input_state_dcs_intermediate = {
+static const struct input_state input_state_dcs_intermediate = {
 	"dcs_intermediate",
 	NULL, NULL,
 	input_state_dcs_intermediate_table
 };
 
 /* dcs_handler state definition. */
-const struct input_state input_state_dcs_handler = {
+static const struct input_state input_state_dcs_handler = {
 	"dcs_handler",
 	NULL, NULL,
 	input_state_dcs_handler_table
 };
 
 /* dcs_escape state definition. */
-const struct input_state input_state_dcs_escape = {
+static const struct input_state input_state_dcs_escape = {
 	"dcs_escape",
 	NULL, NULL,
 	input_state_dcs_escape_table
 };
 
 /* dcs_ignore state definition. */
-const struct input_state input_state_dcs_ignore = {
+static const struct input_state input_state_dcs_ignore = {
 	"dcs_ignore",
 	NULL, NULL,
 	input_state_dcs_ignore_table
 };
 
 /* osc_string state definition. */
-const struct input_state input_state_osc_string = {
+static const struct input_state input_state_osc_string = {
 	"osc_string",
 	input_enter_osc, input_exit_osc,
 	input_state_osc_string_table
 };
 
 /* apc_string state definition. */
-const struct input_state input_state_apc_string = {
+static const struct input_state input_state_apc_string = {
 	"apc_string",
 	input_enter_apc, input_exit_apc,
 	input_state_apc_string_table
 };
 
 /* rename_string state definition. */
-const struct input_state input_state_rename_string = {
+static const struct input_state input_state_rename_string = {
 	"rename_string",
 	input_enter_rename, input_exit_rename,
 	input_state_rename_string_table
 };
 
 /* consume_st state definition. */
-const struct input_state input_state_consume_st = {
+static const struct input_state input_state_consume_st = {
 	"consume_st",
 	NULL, NULL,
 	input_state_consume_st_table
 };
 
 /* utf8_three state definition. */
-const struct input_state input_state_utf8_three = {
+static const struct input_state input_state_utf8_three = {
 	"utf8_three",
 	NULL, NULL,
 	input_state_utf8_three_table
 };
 
 /* utf8_two state definition. */
-const struct input_state input_state_utf8_two = {
+static const struct input_state input_state_utf8_two = {
 	"utf8_two",
 	NULL, NULL,
 	input_state_utf8_two_table
 };
 
 /* utf8_one state definition. */
-const struct input_state input_state_utf8_one = {
+static const struct input_state input_state_utf8_one = {
 	"utf8_one",
 	NULL, NULL,
 	input_state_utf8_one_table
 };
 
 /* ground state table. */
-const struct input_transition input_state_ground_table[] = {
+static const struct input_transition input_state_ground_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch, NULL },
@@ -456,7 +469,7 @@ const struct input_transition input_state_ground_table[] = {
 };
 
 /* esc_enter state table. */
-const struct input_transition input_state_esc_enter_table[] = {
+static const struct input_transition input_state_esc_enter_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch,  NULL },
@@ -483,7 +496,7 @@ const struct input_transition input_state_esc_enter_table[] = {
 };
 
 /* esc_interm state table. */
-const struct input_transition input_state_esc_intermediate_table[] = {
+static const struct input_transition input_state_esc_intermediate_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch,  NULL },
@@ -497,7 +510,7 @@ const struct input_transition input_state_esc_intermediate_table[] = {
 };
 
 /* csi_enter state table. */
-const struct input_transition input_state_csi_enter_table[] = {
+static const struct input_transition input_state_csi_enter_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch,  NULL },
@@ -515,7 +528,7 @@ const struct input_transition input_state_csi_enter_table[] = {
 };
 
 /* csi_parameter state table. */
-const struct input_transition input_state_csi_parameter_table[] = {
+static const struct input_transition input_state_csi_parameter_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch,  NULL },
@@ -533,7 +546,7 @@ const struct input_transition input_state_csi_parameter_table[] = {
 };
 
 /* csi_intermediate state table. */
-const struct input_transition input_state_csi_intermediate_table[] = {
+static const struct input_transition input_state_csi_intermediate_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch,  NULL },
@@ -548,7 +561,7 @@ const struct input_transition input_state_csi_intermediate_table[] = {
 };
 
 /* csi_ignore state table. */
-const struct input_transition input_state_csi_ignore_table[] = {
+static const struct input_transition input_state_csi_ignore_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, input_c0_dispatch, NULL },
@@ -562,7 +575,7 @@ const struct input_transition input_state_csi_ignore_table[] = {
 };
 
 /* dcs_enter state table. */
-const struct input_transition input_state_dcs_enter_table[] = {
+static const struct input_transition input_state_dcs_enter_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,		  NULL },
@@ -580,7 +593,7 @@ const struct input_transition input_state_dcs_enter_table[] = {
 };
 
 /* dcs_parameter state table. */
-const struct input_transition input_state_dcs_parameter_table[] = {
+static const struct input_transition input_state_dcs_parameter_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,		  NULL },
@@ -598,7 +611,7 @@ const struct input_transition input_state_dcs_parameter_table[] = {
 };
 
 /* dcs_interm state table. */
-const struct input_transition input_state_dcs_intermediate_table[] = {
+static const struct input_transition input_state_dcs_intermediate_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,		  NULL },
@@ -613,7 +626,7 @@ const struct input_transition input_state_dcs_intermediate_table[] = {
 };
 
 /* dcs_handler state table. */
-const struct input_transition input_state_dcs_handler_table[] = {
+static const struct input_transition input_state_dcs_handler_table[] = {
 	/* No INPUT_STATE_ANYWHERE */
 
 	{ 0x00, 0x1a, input_input,  NULL },
@@ -624,7 +637,7 @@ const struct input_transition input_state_dcs_handler_table[] = {
 };
 
 /* dcs_escape state table. */
-const struct input_transition input_state_dcs_escape_table[] = {
+static const struct input_transition input_state_dcs_escape_table[] = {
 	/* No INPUT_STATE_ANYWHERE */
 
 	{ 0x00, 0x5b, input_input,	  &input_state_dcs_handler },
@@ -635,7 +648,7 @@ const struct input_transition input_state_dcs_escape_table[] = {
 };
 
 /* dcs_ignore state table. */
-const struct input_transition input_state_dcs_ignore_table[] = {
+static const struct input_transition input_state_dcs_ignore_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,	    NULL },
@@ -647,7 +660,7 @@ const struct input_transition input_state_dcs_ignore_table[] = {
 };
 
 /* osc_string state table. */
-const struct input_transition input_state_osc_string_table[] = {
+static const struct input_transition input_state_osc_string_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x06, NULL,	    NULL },
@@ -661,7 +674,7 @@ const struct input_transition input_state_osc_string_table[] = {
 };
 
 /* apc_string state table. */
-const struct input_transition input_state_apc_string_table[] = {
+static const struct input_transition input_state_apc_string_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,	    NULL },
@@ -673,7 +686,7 @@ const struct input_transition input_state_apc_string_table[] = {
 };
 
 /* rename_string state table. */
-const struct input_transition input_state_rename_string_table[] = {
+static const struct input_transition input_state_rename_string_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,	    NULL },
@@ -685,7 +698,7 @@ const struct input_transition input_state_rename_string_table[] = {
 };
 
 /* consume_st state table. */
-const struct input_transition input_state_consume_st_table[] = {
+static const struct input_transition input_state_consume_st_table[] = {
 	INPUT_STATE_ANYWHERE,
 
 	{ 0x00, 0x17, NULL,	    NULL },
@@ -697,7 +710,7 @@ const struct input_transition input_state_consume_st_table[] = {
 };
 
 /* utf8_three state table. */
-const struct input_transition input_state_utf8_three_table[] = {
+static const struct input_transition input_state_utf8_three_table[] = {
 	/* No INPUT_STATE_ANYWHERE */
 
 	{ 0x00, 0x7f, NULL,		&input_state_ground },
@@ -708,7 +721,7 @@ const struct input_transition input_state_utf8_three_table[] = {
 };
 
 /* utf8_two state table. */
-const struct input_transition input_state_utf8_two_table[] = {
+static const struct input_transition input_state_utf8_two_table[] = {
 	/* No INPUT_STATE_ANYWHERE */
 
 	{ 0x00, 0x7f, NULL,	      &input_state_ground },
@@ -719,7 +732,7 @@ const struct input_transition input_state_utf8_two_table[] = {
 };
 
 /* utf8_one state table. */
-const struct input_transition input_state_utf8_one_table[] = {
+static const struct input_transition input_state_utf8_one_table[] = {
 	/* No INPUT_STATE_ANYWHERE */
 
 	{ 0x00, 0x7f, NULL,		&input_state_ground },
@@ -730,7 +743,7 @@ const struct input_transition input_state_utf8_one_table[] = {
 };
 
 /* Input table compare. */
-int
+static int
 input_table_compare(const void *key, const void *value)
 {
 	const struct input_ctx		*ictx = key;
@@ -742,7 +755,7 @@ input_table_compare(const void *key, const void *value)
 }
 
 /* Reset cell state to default. */
-void
+static void
 input_reset_cell(struct input_ctx *ictx)
 {
 	memcpy(&ictx->cell.cell, &grid_default_cell, sizeof ictx->cell.cell);
@@ -779,7 +792,7 @@ input_free(struct window_pane *wp)
 	free(ictx->input_buf);
 	evbuffer_free(ictx->since_ground);
 
-	free (ictx);
+	free(ictx);
 	wp->ictx = NULL;
 }
 
@@ -821,7 +834,7 @@ input_pending(struct window_pane *wp)
 }
 
 /* Change input state. */
-void
+static void
 input_set_state(struct window_pane *wp, const struct input_transition *itr)
 {
 	struct input_ctx	*ictx = wp->ictx;
@@ -861,9 +874,10 @@ input_parse(struct window_pane *wp)
 
 	buf = EVBUFFER_DATA(evb);
 	len = EVBUFFER_LENGTH(evb);
-	notify_input(wp, evb);
 	off = 0;
 
+	notify_input(wp, evb);
+
 	log_debug("%s: %%%u %s, %zu bytes: %.*s", __func__, wp->id,
 	    ictx->state->name, len, (int)len, buf);
 
@@ -883,6 +897,16 @@ input_parse(struct window_pane *wp)
 			fatalx("no transition from state");
 		}
 
+		/*
+		 * Any state except print stops the current collection. This is
+		 * an optimization to avoid checking if the attributes have
+		 * changed for every character. It will stop unnecessarily for
+		 * sequences that don't make a terminal change, but they should
+		 * be the minority.
+		 */
+		if (itr->handler != input_print)
+			screen_write_collect_end(&ictx->ctx);
+
 		/*
 		 * Execute the handler, if any. Don't switch state if it
 		 * returns non-zero.
@@ -906,9 +930,8 @@ input_parse(struct window_pane *wp)
 }
 
 /* Split the parameter list (if any). */
-int
+static int
 input_split(struct input_ctx *ictx)
-
 {
 	const char	*errstr;
 	char		*ptr, *out;
@@ -937,7 +960,7 @@ input_split(struct input_ctx *ictx)
 }
 
 /* Get an argument or return default value. */
-int
+static int
 input_get(struct input_ctx *ictx, u_int validx, int minval, int defval)
 {
 	int	retval;
@@ -954,7 +977,7 @@ input_get(struct input_ctx *ictx, u_int validx, int minval, int defval)
 }
 
 /* Reply to terminal query. */
-void
+static void
 input_reply(struct input_ctx *ictx, const char *fmt, ...)
 {
 	va_list	ap;
@@ -969,7 +992,7 @@ input_reply(struct input_ctx *ictx, const char *fmt, ...)
 }
 
 /* Clear saved state. */
-void
+static void
 input_clear(struct input_ctx *ictx)
 {
 	*ictx->interm_buf = '\0';
@@ -985,7 +1008,7 @@ input_clear(struct input_ctx *ictx)
 }
 
 /* Reset for ground state. */
-void
+static void
 input_ground(struct input_ctx *ictx)
 {
 	evbuffer_drain(ictx->since_ground, EVBUFFER_LENGTH(ictx->since_ground));
@@ -997,7 +1020,7 @@ input_ground(struct input_ctx *ictx)
 }
 
 /* Output this character to the screen. */
-int
+static int
 input_print(struct input_ctx *ictx)
 {
 	int	set;
@@ -1009,7 +1032,7 @@ input_print(struct input_ctx *ictx)
 		ictx->cell.cell.attr &= ~GRID_ATTR_CHARSET;
 
 	utf8_set(&ictx->cell.cell.data, ictx->ch);
-	screen_write_cell(&ictx->ctx, &ictx->cell.cell);
+	screen_write_collect_add(&ictx->ctx, &ictx->cell.cell);
 
 	ictx->cell.cell.attr &= ~GRID_ATTR_CHARSET;
 
@@ -1017,7 +1040,7 @@ input_print(struct input_ctx *ictx)
 }
 
 /* Collect intermediate string. */
-int
+static int
 input_intermediate(struct input_ctx *ictx)
 {
 	if (ictx->interm_len == (sizeof ictx->interm_buf) - 1)
@@ -1031,7 +1054,7 @@ input_intermediate(struct input_ctx *ictx)
 }
 
 /* Collect parameter string. */
-int
+static int
 input_parameter(struct input_ctx *ictx)
 {
 	if (ictx->param_len == (sizeof ictx->param_buf) - 1)
@@ -1045,7 +1068,7 @@ input_parameter(struct input_ctx *ictx)
 }
 
 /* Collect input string. */
-int
+static int
 input_input(struct input_ctx *ictx)
 {
 	size_t available;
@@ -1067,7 +1090,7 @@ input_input(struct input_ctx *ictx)
 }
 
 /* Execute C0 control sequence. */
-int
+static int
 input_c0_dispatch(struct input_ctx *ictx)
 {
 	struct screen_write_ctx	*sctx = &ictx->ctx;
@@ -1120,7 +1143,7 @@ input_c0_dispatch(struct input_ctx *ictx)
 }
 
 /* Execute escape sequence. */
-int
+static int
 input_esc_dispatch(struct input_ctx *ictx)
 {
 	struct screen_write_ctx		*sctx = &ictx->ctx;
@@ -1140,6 +1163,7 @@ input_esc_dispatch(struct input_ctx *ictx)
 
 	switch (entry->type) {
 	case INPUT_ESC_RIS:
+		window_pane_reset_palette(ictx->wp);
 		input_reset_cell(ictx);
 		screen_write_reset(sctx);
 		break;
@@ -1187,13 +1211,16 @@ input_esc_dispatch(struct input_ctx *ictx)
 	case INPUT_ESC_SCSG1_OFF:
 		ictx->cell.g1set = 0;
 		break;
+	case INPUT_ESC_ST:
+		/* ST terminates OSC but the state transition already did it. */
+		break;
 	}
 
 	return (0);
 }
 
 /* Execute control sequence. */
-int
+static int
 input_csi_dispatch(struct input_ctx *ictx)
 {
 	struct screen_write_ctx	       *sctx = &ictx->ctx;
@@ -1204,11 +1231,13 @@ input_csi_dispatch(struct input_ctx *ictx)
 
 	if (ictx->flags & INPUT_DISCARD)
 		return (0);
-	if (input_split(ictx) != 0)
-		return (0);
+
 	log_debug("%s: '%c' \"%s\" \"%s\"",
 	    __func__, ictx->ch, ictx->interm_buf, ictx->param_buf);
 
+	if (input_split(ictx) != 0)
+		return (0);
+
 	entry = bsearch(ictx, input_csi_table, nitems(input_csi_table),
 	    sizeof input_csi_table[0], input_table_compare);
 	if (entry == NULL) {
@@ -1279,10 +1308,12 @@ input_csi_dispatch(struct input_ctx *ictx)
 		}
 		break;
 	case INPUT_CSI_ECH:
-		screen_write_clearcharacter(sctx, input_get(ictx, 0, 1, 1));
+		screen_write_clearcharacter(sctx, input_get(ictx, 0, 1, 1),
+		    ictx->cell.cell.bg);
 		break;
 	case INPUT_CSI_DCH:
-		screen_write_deletecharacter(sctx, input_get(ictx, 0, 1, 1));
+		screen_write_deletecharacter(sctx, input_get(ictx, 0, 1, 1),
+		    ictx->cell.cell.bg);
 		break;
 	case INPUT_CSI_DECSTBM:
 		n = input_get(ictx, 0, 1, 1);
@@ -1290,7 +1321,8 @@ input_csi_dispatch(struct input_ctx *ictx)
 		screen_write_scrollregion(sctx, n - 1, m - 1);
 		break;
 	case INPUT_CSI_DL:
-		screen_write_deleteline(sctx, input_get(ictx, 0, 1, 1));
+		screen_write_deleteline(sctx, input_get(ictx, 0, 1, 1),
+		    ictx->cell.cell.bg);
 		break;
 	case INPUT_CSI_DSR:
 		switch (input_get(ictx, 0, 0, 0)) {
@@ -1308,13 +1340,13 @@ input_csi_dispatch(struct input_ctx *ictx)
 	case INPUT_CSI_ED:
 		switch (input_get(ictx, 0, 0, 0)) {
 		case 0:
-			screen_write_clearendofscreen(sctx);
+			screen_write_clearendofscreen(sctx, ictx->cell.cell.bg);
 			break;
 		case 1:
-			screen_write_clearstartofscreen(sctx);
+			screen_write_clearstartofscreen(sctx, ictx->cell.cell.bg);
 			break;
 		case 2:
-			screen_write_clearscreen(sctx);
+			screen_write_clearscreen(sctx, ictx->cell.cell.bg);
 			break;
 		case 3:
 			switch (input_get(ictx, 1, 0, 0)) {
@@ -1335,13 +1367,13 @@ input_csi_dispatch(struct input_ctx *ictx)
 	case INPUT_CSI_EL:
 		switch (input_get(ictx, 0, 0, 0)) {
 		case 0:
-			screen_write_clearendofline(sctx);
+			screen_write_clearendofline(sctx, ictx->cell.cell.bg);
 			break;
 		case 1:
-			screen_write_clearstartofline(sctx);
+			screen_write_clearstartofline(sctx, ictx->cell.cell.bg);
 			break;
 		case 2:
-			screen_write_clearline(sctx);
+			screen_write_clearline(sctx, ictx->cell.cell.bg);
 			break;
 		default:
 			log_debug("%s: unknown '%c'", __func__, ictx->ch);
@@ -1353,10 +1385,12 @@ input_csi_dispatch(struct input_ctx *ictx)
 		screen_write_cursormove(sctx, n - 1, s->cy);
 		break;
 	case INPUT_CSI_ICH:
-		screen_write_insertcharacter(sctx, input_get(ictx, 0, 1, 1));
+		screen_write_insertcharacter(sctx, input_get(ictx, 0, 1, 1),
+		    ictx->cell.cell.bg);
 		break;
 	case INPUT_CSI_IL:
-		screen_write_insertline(sctx, input_get(ictx, 0, 1, 1));
+		screen_write_insertline(sctx, input_get(ictx, 0, 1, 1),
+		    ictx->cell.cell.bg);
 		break;
 	case INPUT_CSI_RCP:
 		memcpy(&ictx->cell, &ictx->old_cell, sizeof ictx->cell);
@@ -1382,6 +1416,9 @@ input_csi_dispatch(struct input_ctx *ictx)
 	case INPUT_CSI_SM_PRIVATE:
 		input_csi_dispatch_sm_private(ictx);
 		break;
+	case INPUT_CSI_SU:
+		screen_write_scrollup(sctx, input_get(ictx, 0, 1, 1));
+		break;
 	case INPUT_CSI_TBC:
 		switch (input_get(ictx, 0, 0, 0)) {
 		case 0:
@@ -1410,7 +1447,7 @@ input_csi_dispatch(struct input_ctx *ictx)
 }
 
 /* Handle CSI RM. */
-void
+static void
 input_csi_dispatch_rm(struct input_ctx *ictx)
 {
 	u_int	i;
@@ -1431,7 +1468,7 @@ input_csi_dispatch_rm(struct input_ctx *ictx)
 }
 
 /* Handle CSI private RM. */
-void
+static void
 input_csi_dispatch_rm_private(struct input_ctx *ictx)
 {
 	struct window_pane	*wp = ictx->wp;
@@ -1444,7 +1481,8 @@ input_csi_dispatch_rm_private(struct input_ctx *ictx)
 			break;
 		case 3:		/* DECCOLM */
 			screen_write_cursormove(&ictx->ctx, 0, 0);
-			screen_write_clearscreen(&ictx->ctx);
+			screen_write_clearscreen(&ictx->ctx,
+			    ictx->cell.cell.bg);
 			break;
 		case 7:		/* DECAWM */
 			screen_write_mode_clear(&ictx->ctx, MODE_WRAP);
@@ -1458,6 +1496,7 @@ input_csi_dispatch_rm_private(struct input_ctx *ictx)
 		case 1000:
 		case 1001:
 		case 1002:
+		case 1003:
 			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
 			break;
 		case 1004:
@@ -1487,7 +1526,7 @@ input_csi_dispatch_rm_private(struct input_ctx *ictx)
 }
 
 /* Handle CSI SM. */
-void
+static void
 input_csi_dispatch_sm(struct input_ctx *ictx)
 {
 	u_int	i;
@@ -1508,7 +1547,7 @@ input_csi_dispatch_sm(struct input_ctx *ictx)
 }
 
 /* Handle CSI private SM. */
-void
+static void
 input_csi_dispatch_sm_private(struct input_ctx *ictx)
 {
 	struct window_pane	*wp = ictx->wp;
@@ -1521,7 +1560,8 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 			break;
 		case 3:		/* DECCOLM */
 			screen_write_cursormove(&ictx->ctx, 0, 0);
-			screen_write_clearscreen(&ictx->ctx);
+			screen_write_clearscreen(&ictx->ctx,
+			    ictx->cell.cell.bg);
 			break;
 		case 7:		/* DECAWM */
 			screen_write_mode_set(&ictx->ctx, MODE_WRAP);
@@ -1540,6 +1580,10 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
 			screen_write_mode_set(&ictx->ctx, MODE_MOUSE_BUTTON);
 			break;
+		case 1003:
+			screen_write_mode_clear(&ictx->ctx, ALL_MOUSE_MODES);
+			screen_write_mode_set(&ictx->ctx, MODE_MOUSE_ALL);
+			break;
 		case 1004:
 			if (ictx->ctx.s->mode & MODE_FOCUSON)
 				break;
@@ -1570,7 +1614,7 @@ input_csi_dispatch_sm_private(struct input_ctx *ictx)
 }
 
 /* Handle CSI window operations. */
-void
+static void
 input_csi_dispatch_winops(struct input_ctx *ictx)
 {
 	struct window_pane	*wp = ictx->wp;
@@ -1619,7 +1663,7 @@ input_csi_dispatch_winops(struct input_ctx *ictx)
 }
 
 /* Handle CSI SGR for 256 colours. */
-void
+static void
 input_csi_dispatch_sgr_256(struct input_ctx *ictx, int fgbg, u_int *i)
 {
 	struct grid_cell	*gc = &ictx->cell.cell;
@@ -1641,7 +1685,7 @@ input_csi_dispatch_sgr_256(struct input_ctx *ictx, int fgbg, u_int *i)
 }
 
 /* Handle CSI SGR for RGB colours. */
-void
+static void
 input_csi_dispatch_sgr_rgb(struct input_ctx *ictx, int fgbg, u_int *i)
 {
 	struct grid_cell	*gc = &ictx->cell.cell;
@@ -1667,7 +1711,7 @@ input_csi_dispatch_sgr_rgb(struct input_ctx *ictx, int fgbg, u_int *i)
 }
 
 /* Handle CSI SGR. */
-void
+static void
 input_csi_dispatch_sgr(struct input_ctx *ictx)
 {
 	struct grid_cell	*gc = &ictx->cell.cell;
@@ -1721,6 +1765,9 @@ input_csi_dispatch_sgr(struct input_ctx *ictx)
 		case 8:
 			gc->attr |= GRID_ATTR_HIDDEN;
 			break;
+		case 9:
+			gc->attr |= GRID_ATTR_STRIKETHROUGH;
+			break;
 		case 22:
 			gc->attr &= ~(GRID_ATTR_BRIGHT|GRID_ATTR_DIM);
 			break;
@@ -1736,6 +1783,12 @@ input_csi_dispatch_sgr(struct input_ctx *ictx)
 		case 27:
 			gc->attr &= ~GRID_ATTR_REVERSE;
 			break;
+		case 28:
+			gc->attr &= ~GRID_ATTR_HIDDEN;
+			break;
+		case 29:
+			gc->attr &= ~GRID_ATTR_STRIKETHROUGH;
+			break;
 		case 30:
 		case 31:
 		case 32:
@@ -1787,7 +1840,7 @@ input_csi_dispatch_sgr(struct input_ctx *ictx)
 }
 
 /* DCS terminator (ST) received. */
-int
+static int
 input_dcs_dispatch(struct input_ctx *ictx)
 {
 	const char	prefix[] = "tmux;";
@@ -1809,7 +1862,7 @@ input_dcs_dispatch(struct input_ctx *ictx)
 }
 
 /* OSC string started. */
-void
+static void
 input_enter_osc(struct input_ctx *ictx)
 {
 	log_debug("%s", __func__);
@@ -1818,7 +1871,7 @@ input_enter_osc(struct input_ctx *ictx)
 }
 
 /* OSC terminator (ST) received. */
-void
+static void
 input_exit_osc(struct input_ctx *ictx)
 {
 	u_char	*p = ictx->input_buf;
@@ -1843,10 +1896,19 @@ input_exit_osc(struct input_ctx *ictx)
 		screen_set_title(ictx->ctx.s, p);
 		server_status_window(ictx->wp->window);
 		break;
+	case 4:
+		input_osc_4(ictx->wp, p);
+		break;
+	case 52:
+		input_osc_52(ictx->wp, p);
+		break;
 	case 12:
 		if (*p != '?') /* ? is colour request */
 			screen_set_cursor_colour(ictx->ctx.s, p);
 		break;
+	case 104:
+		input_osc_104(ictx->wp, p);
+		break;
 	case 112:
 		if (*p == '\0') /* no arguments allowed */
 			screen_set_cursor_colour(ictx->ctx.s, "");
@@ -1858,7 +1920,7 @@ input_exit_osc(struct input_ctx *ictx)
 }
 
 /* APC string started. */
-void
+static void
 input_enter_apc(struct input_ctx *ictx)
 {
 	log_debug("%s", __func__);
@@ -1867,7 +1929,7 @@ input_enter_apc(struct input_ctx *ictx)
 }
 
 /* APC terminator (ST) received. */
-void
+static void
 input_exit_apc(struct input_ctx *ictx)
 {
 	if (ictx->flags & INPUT_DISCARD)
@@ -1879,7 +1941,7 @@ input_exit_apc(struct input_ctx *ictx)
 }
 
 /* Rename string started. */
-void
+static void
 input_enter_rename(struct input_ctx *ictx)
 {
 	log_debug("%s", __func__);
@@ -1888,7 +1950,7 @@ input_enter_rename(struct input_ctx *ictx)
 }
 
 /* Rename terminator (ST) received. */
-void
+static void
 input_exit_rename(struct input_ctx *ictx)
 {
 	if (ictx->flags & INPUT_DISCARD)
@@ -1904,7 +1966,7 @@ input_exit_rename(struct input_ctx *ictx)
 }
 
 /* Open UTF-8 character. */
-int
+static int
 input_utf8_open(struct input_ctx *ictx)
 {
 	struct utf8_data	*ud = &ictx->utf8data;
@@ -1918,7 +1980,7 @@ input_utf8_open(struct input_ctx *ictx)
 }
 
 /* Append to UTF-8 character. */
-int
+static int
 input_utf8_add(struct input_ctx *ictx)
 {
 	struct utf8_data	*ud = &ictx->utf8data;
@@ -1932,7 +1994,7 @@ input_utf8_add(struct input_ctx *ictx)
 }
 
 /* Close UTF-8 string. */
-int
+static int
 input_utf8_close(struct input_ctx *ictx)
 {
 	struct utf8_data	*ud = &ictx->utf8data;
@@ -1950,7 +2012,107 @@ input_utf8_close(struct input_ctx *ictx)
 	    (int)ud->size, ud->data, ud->width);
 
 	utf8_copy(&ictx->cell.cell.data, ud);
-	screen_write_cell(&ictx->ctx, &ictx->cell.cell);
+	screen_write_collect_add(&ictx->ctx, &ictx->cell.cell);
 
 	return (0);
 }
+
+/* Handle the OSC 4 sequence for setting (multiple) palette entries. */
+static void
+input_osc_4(struct window_pane *wp, const char *p)
+{
+	char	*copy, *s, *next = NULL;
+	long	 idx;
+	u_int	 r, g, b;
+
+	copy = s = xstrdup(p);
+	while (s != NULL && *s != '\0') {
+		idx = strtol(s, &next, 10);
+		if (*next++ != ';')
+			goto bad;
+		if (idx < 0 || idx >= 0x100)
+			goto bad;
+
+		s = strsep(&next, ";");
+		if (sscanf(s, "rgb:%2x/%2x/%2x", &r, &g, &b) != 3) {
+			s = next;
+			continue;
+		}
+
+		window_pane_set_palette(wp, idx, colour_join_rgb(r, g, b));
+		s = next;
+	}
+
+	free(copy);
+	return;
+
+bad:
+	log_debug("bad OSC 4: %s", p);
+	free(copy);
+}
+
+/* Handle the OSC 52 sequence for setting the clipboard. */
+static void
+input_osc_52(struct window_pane *wp, const char *p)
+{
+	char			*end;
+	size_t			 len;
+	u_char			*out;
+	int			 outlen;
+	struct screen_write_ctx	 ctx;
+
+	if ((end = strchr(p, ';')) == NULL)
+		return;
+	end++;
+	if (*end == '\0')
+		return;
+
+	len = (strlen(end) / 4) * 3;
+	if (len == 0)
+		return;
+
+	out = xmalloc(len);
+	if ((outlen = b64_pton(end, out, len)) == -1) {
+		free(out);
+		return;
+	}
+
+	if (options_get_number(global_options, "set-clipboard")) {
+		screen_write_start(&ctx, wp, NULL);
+		screen_write_setselection(&ctx, out, outlen);
+		screen_write_stop(&ctx);
+	}
+	paste_add(out, outlen);
+}
+
+/* Handle the OSC 104 sequence for unsetting (multiple) palette entries. */
+static void
+input_osc_104(struct window_pane *wp, const char *p)
+{
+	char	*copy, *s;
+	long	idx;
+
+	if (*p == '\0') {
+		window_pane_reset_palette(wp);
+		return;
+	}
+
+	copy = s = xstrdup(p);
+	while (*s != '\0') {
+		idx = strtol(s, &s, 10);
+		if (*s != '\0' && *s != ';')
+			goto bad;
+		if (idx < 0 || idx >= 0x100)
+			goto bad;
+
+		window_pane_unset_palette(wp, idx);
+		if (*s == ';')
+			s++;
+	}
+	free(copy);
+	return;
+
+bad:
+	log_debug("bad OSC 104: %s", p);
+	free(copy);
+}
