@@ -26,26 +26,8 @@
 
 #include "tmux.h"
 
-struct session *server_next_session(struct session *);
-void		server_callback_identify(int, short, void *);
-
-void
-server_fill_environ(struct session *s, struct environ *env)
-{
-	char	*term;
-	u_int	 idx;
-	long	 pid;
-
-	if (s != NULL) {
-		term = options_get_string(global_options, "default-terminal");
-		environ_set(env, "TERM", "%s", term);
-
-		idx = s->id;
-	} else
-		idx = (u_int)-1;
-	pid = getpid();
-	environ_set(env, "TMUX", "%s,%ld,%u", socket_path, pid, idx);
-}
+static struct session	*server_next_session(struct session *);
+static void		 server_destroy_session_group(struct session *);
 
 void
 server_redraw_client(struct client *c)
@@ -75,7 +57,7 @@ server_redraw_session_group(struct session *s)
 {
 	struct session_group	*sg;
 
-	if ((sg = session_group_find(s)) == NULL)
+	if ((sg = session_group_contains(s)) == NULL)
 		server_redraw_session(s);
 	else {
 		TAILQ_FOREACH(s, &sg->sessions, gentry)
@@ -99,7 +81,7 @@ server_status_session_group(struct session *s)
 {
 	struct session_group	*sg;
 
-	if ((sg = session_group_find(s)) == NULL)
+	if ((sg = session_group_contains(s)) == NULL)
 		server_status_session(s);
 	else {
 		TAILQ_FOREACH(s, &sg->sessions, gentry)
@@ -116,7 +98,6 @@ server_redraw_window(struct window *w)
 		if (c->session != NULL && c->session->curw->window == w)
 			server_redraw_client(c);
 	}
-	w->flags |= WINDOW_REDRAW;
 }
 
 void
@@ -217,7 +198,7 @@ server_kill_window(struct window *w)
 		}
 
 		if (options_get_number(s->options, "renumber-windows")) {
-			if ((sg = session_group_find(s)) != NULL) {
+			if ((sg = session_group_contains(s)) != NULL) {
 				TAILQ_FOREACH(target_s, &sg->sessions, gentry)
 					session_renumber_windows(target_s);
 			} else
@@ -235,8 +216,8 @@ server_link_window(struct session *src, struct winlink *srcwl,
 	struct winlink		*dstwl;
 	struct session_group	*srcsg, *dstsg;
 
-	srcsg = session_group_find(src);
-	dstsg = session_group_find(dst);
+	srcsg = session_group_contains(src);
+	dstsg = session_group_contains(dst);
 	if (src != dst && srcsg != NULL && dstsg != NULL && srcsg == dstsg) {
 		xasprintf(cause, "sessions are grouped");
 		return (-1);
@@ -255,7 +236,8 @@ server_link_window(struct session *src, struct winlink *srcwl,
 			 * Can't use session_detach as it will destroy session
 			 * if this makes it empty.
 			 */
-			notify_window_unlinked(dst, dstwl->window);
+			notify_session_window("window-unlinked", dst,
+			    dstwl->window);
 			dstwl->flags &= ~WINLINK_ALERTFLAGS;
 			winlink_stack_remove(&dst->lastw, dstwl);
 			winlink_remove(&dst->windows, dstwl);
@@ -291,13 +273,12 @@ server_unlink_window(struct session *s, struct winlink *wl)
 }
 
 void
-server_destroy_pane(struct window_pane *wp, int hooks)
+server_destroy_pane(struct window_pane *wp, int notify)
 {
 	struct window		*w = wp->window;
 	int			 old_fd;
 	struct screen_write_ctx	 ctx;
 	struct grid_cell	 gc;
-	struct cmd_find_state	 fs;
 
 	old_fd = wp->fd;
 	if (wp->fd != -1) {
@@ -312,6 +293,10 @@ server_destroy_pane(struct window_pane *wp, int hooks)
 	if (options_get_number(w->options, "remain-on-exit")) {
 		if (old_fd == -1)
 			return;
+
+		if (notify)
+			notify_pane("pane-died", wp);
+
 		screen_write_start(&ctx, wp, &wp->base);
 		screen_write_scrollregion(&ctx, 0, screen_size_y(ctx.s) - 1);
 		screen_write_cursormove(&ctx, 0, screen_size_y(ctx.s) - 1);
@@ -322,31 +307,29 @@ server_destroy_pane(struct window_pane *wp, int hooks)
 		screen_write_stop(&ctx);
 		wp->flags |= PANE_REDRAW;
 
-		if (hooks && cmd_find_from_pane(&fs, wp) == 0)
-			hooks_run(hooks_get(fs.s), NULL, &fs, "pane-died");
 		return;
 	}
 
+	if (notify)
+		notify_pane("pane-exited", wp);
+
 	server_unzoom_window(w);
 	layout_close_pane(wp);
 	window_remove_pane(w, wp);
 
-	if (hooks && cmd_find_from_window(&fs, w) == 0)
-		hooks_run(hooks_get(fs.s), NULL, &fs, "pane-exited");
-
 	if (TAILQ_EMPTY(&w->panes))
 		server_kill_window(w);
 	else
 		server_redraw_window(w);
 }
 
-void
+static void
 server_destroy_session_group(struct session *s)
 {
 	struct session_group	*sg;
 	struct session		*s1;
 
-	if ((sg = session_group_find(s)) == NULL)
+	if ((sg = session_group_contains(s)) == NULL)
 		server_destroy_session(s);
 	else {
 		TAILQ_FOREACH_SAFE(s, &sg->sessions, gentry, s1) {
@@ -356,7 +339,7 @@ server_destroy_session_group(struct session *s)
 	}
 }
 
-struct session *
+static struct session *
 server_next_session(struct session *s)
 {
 	struct session *s_loop, *s_out;
@@ -394,7 +377,7 @@ server_destroy_session(struct session *s)
 			c->session = s_new;
 			server_client_set_key_table(c, NULL);
 			status_timer_start(c);
-			notify_attached_session_changed(c);
+			notify_client("client-session-changed", c);
 			session_update_activity(s_new, NULL);
 			gettimeofday(&s_new->last_attached_time, NULL);
 			server_redraw_client(c);
@@ -421,46 +404,6 @@ server_check_unattached(void)
 	}
 }
 
-void
-server_set_identify(struct client *c)
-{
-	struct timeval	tv;
-	int		delay;
-
-	delay = options_get_number(c->session->options, "display-panes-time");
-	tv.tv_sec = delay / 1000;
-	tv.tv_usec = (delay % 1000) * 1000L;
-
-	if (event_initialized(&c->identify_timer))
-		evtimer_del(&c->identify_timer);
-	evtimer_set(&c->identify_timer, server_callback_identify, c);
-	evtimer_add(&c->identify_timer, &tv);
-
-	c->flags |= CLIENT_IDENTIFY;
-	c->tty.flags |= (TTY_FREEZE|TTY_NOCURSOR);
-	server_redraw_client(c);
-}
-
-void
-server_clear_identify(struct client *c, struct window_pane *wp)
-{
-	if (~c->flags & CLIENT_IDENTIFY)
-		return;
-	c->flags &= ~CLIENT_IDENTIFY;
-
-	if (c->identify_callback != NULL)
-		c->identify_callback(c, wp);
-
-	c->tty.flags &= ~(TTY_FREEZE|TTY_NOCURSOR);
-	server_redraw_client(c);
-}
-
-void
-server_callback_identify(__unused int fd, __unused short events, void *data)
-{
-	server_clear_identify(data, NULL);
-}
-
 /* Set stdin callback. */
 int
 server_set_stdin_callback(struct client *c, void (*cb)(struct client *, int,
