@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2007 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2007 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -26,16 +26,23 @@
 
 #include "tmux.h"
 
-struct sessions	sessions;
-u_int		next_session_id;
-struct session_groups session_groups;
+struct sessions		sessions;
+static u_int		next_session_id;
+struct session_groups	session_groups;
 
-void	session_free(int, short, void *);
+static void	session_free(int, short, void *);
 
-void	session_lock_timer(int, short, void *);
+static void	session_lock_timer(int, short, void *);
 
-struct winlink *session_next_alert(struct winlink *);
-struct winlink *session_previous_alert(struct winlink *);
+static struct winlink *session_next_alert(struct winlink *);
+static struct winlink *session_previous_alert(struct winlink *);
+
+static void	session_group_remove(struct session *);
+static u_int	session_group_count(struct session_group *);
+static void	session_group_synchronize1(struct session *, struct session *);
+
+static u_int	session_group_count(struct session_group *);
+static void	session_group_synchronize1(struct session *, struct session *);
 
 RB_GENERATE(sessions, session, entry, session_cmp);
 
@@ -45,6 +52,14 @@ session_cmp(struct session *s1, struct session *s2)
 	return (strcmp(s1->name, s2->name));
 }
 
+RB_GENERATE(session_groups, session_group, entry, session_group_cmp);
+
+int
+session_group_cmp(struct session_group *s1, struct session_group *s2)
+{
+	return (strcmp(s1->name, s2->name));
+}
+
 /*
  * Find if session is still alive. This is true if it is still on the global
  * sessions list.
@@ -102,9 +117,9 @@ session_find_by_id(u_int id)
 
 /* Create a new session. */
 struct session *
-session_create(const char *name, int argc, char **argv, const char *path,
-    int cwd, struct environ *env, struct termios *tio, int idx, u_int sx,
-    u_int sy, char **cause)
+session_create(const char *prefix, const char *name, int argc, char **argv,
+    const char *path, const char *cwd, struct environ *env, struct termios *tio,
+    int idx, u_int sx, u_int sy, char **cause)
 {
 	struct session	*s;
 	struct winlink	*wl;
@@ -113,16 +128,20 @@ session_create(const char *name, int argc, char **argv, const char *path,
 	s->references = 1;
 	s->flags = 0;
 
-	s->cwd = dup(cwd);
+	s->cwd = xstrdup(cwd);
 
 	s->curw = NULL;
 	TAILQ_INIT(&s->lastw);
 	RB_INIT(&s->windows);
 
-	options_init(&s->options, &global_s_options);
-	environ_init(&s->environ);
+	s->environ = environ_create();
 	if (env != NULL)
-		environ_copy(env, &s->environ);
+		environ_copy(env, s->environ);
+
+	s->options = options_create(global_s_options);
+	s->hooks = hooks_create(global_hooks);
+
+	status_update_saved(s);
 
 	s->tio = NULL;
 	if (tio != NULL) {
@@ -130,10 +149,6 @@ session_create(const char *name, int argc, char **argv, const char *path,
 		memcpy(s->tio, tio, sizeof *s->tio);
 	}
 
-	if (gettimeofday(&s->creation_time, NULL) != 0)
-		fatal("gettimeofday failed");
-	session_update_activity(s, &s->creation_time);
-
 	s->sx = sx;
 	s->sy = sy;
 
@@ -145,11 +160,16 @@ session_create(const char *name, int argc, char **argv, const char *path,
 		do {
 			s->id = next_session_id++;
 			free(s->name);
-			xasprintf(&s->name, "%u", s->id);
+			if (prefix != NULL)
+				xasprintf(&s->name, "%s-%u", prefix, s->id);
+			else
+				xasprintf(&s->name, "%u", s->id);
 		} while (RB_FIND(sessions, &sessions, s) != NULL);
 	}
 	RB_INSERT(sessions, &sessions, s);
 
+	log_debug("new session %s $%u", s->name, s->id);
+
 	if (gettimeofday(&s->creation_time, NULL) != 0)
 		fatal("gettimeofday failed");
 	session_update_activity(s, &s->creation_time);
@@ -164,31 +184,43 @@ session_create(const char *name, int argc, char **argv, const char *path,
 	}
 
 	log_debug("session %s created", s->name);
-	notify_session_created(s);
 
 	return (s);
 }
 
+/* Add a reference to a session. */
+void
+session_add_ref(struct session *s, const char *from)
+{
+	s->references++;
+	log_debug("%s: %s %s, now %d", __func__, s->name, from, s->references);
+}
+
 /* Remove a reference from a session. */
 void
-session_unref(struct session *s)
+session_remove_ref(struct session *s, const char *from)
 {
-	log_debug("session %s has %d references", s->name, s->references);
-
 	s->references--;
+	log_debug("%s: %s %s, now %d", __func__, s->name, from, s->references);
+
 	if (s->references == 0)
 		event_once(-1, EV_TIMEOUT, session_free, s, NULL);
 }
 
 /* Free session. */
-void
-session_free(unused int fd, unused short events, void *arg)
+static void
+session_free(__unused int fd, __unused short events, void *arg)
 {
 	struct session	*s = arg;
 
 	log_debug("session %s freed (%d references)", s->name, s->references);
 
 	if (s->references == 0) {
+		environ_free(s->environ);
+
+		options_free(s->options);
+		hooks_free(s->hooks);
+
 		free(s->name);
 		free(s);
 	}
@@ -201,9 +233,10 @@ session_destroy(struct session *s)
 	struct winlink	*wl;
 
 	log_debug("session %s destroyed", s->name);
+	s->curw = NULL;
 
 	RB_REMOVE(sessions, &sessions, s);
-	notify_session_closed(s);
+	notify_session("session-closed", s);
 
 	free(s->tio);
 
@@ -211,20 +244,18 @@ session_destroy(struct session *s)
 		event_del(&s->lock_timer);
 
 	session_group_remove(s);
-	environ_free(&s->environ);
-	options_free(&s->options);
 
 	while (!TAILQ_EMPTY(&s->lastw))
 		winlink_stack_remove(&s->lastw, TAILQ_FIRST(&s->lastw));
 	while (!RB_EMPTY(&s->windows)) {
 		wl = RB_ROOT(&s->windows);
-		notify_window_unlinked(s, wl->window);
+		notify_session_window("window-unlinked", s, wl->window);
 		winlink_remove(&s->windows, wl);
 	}
 
-	close(s->cwd);
+	free((void *)s->cwd);
 
-	session_unref(s);
+	session_remove_ref(s, __func__);
 }
 
 /* Check a session name is valid: not empty and no colons or periods. */
@@ -235,8 +266,8 @@ session_check_name(const char *name)
 }
 
 /* Lock session if it has timed out. */
-void
-session_lock_timer(unused int fd, unused short events, void *arg)
+static void
+session_lock_timer(__unused int fd, __unused short events, void *arg)
 {
 	struct session	*s = arg;
 
@@ -263,6 +294,10 @@ session_update_activity(struct session *s, struct timeval *from)
 	else
 		memcpy(&s->activity_time, from, sizeof s->activity_time);
 
+	log_debug("session %s activity %lld.%06d (last %lld.%06d)", s->name,
+	    (long long)s->activity_time.tv_sec, (int)s->activity_time.tv_usec,
+	    (long long)last->tv_sec, (int)last->tv_usec);
+
 	if (evtimer_initialized(&s->lock_timer))
 		evtimer_del(&s->lock_timer);
 	else
@@ -270,7 +305,7 @@ session_update_activity(struct session *s, struct timeval *from)
 
 	if (~s->flags & SESSION_UNATTACHED) {
 		timerclear(&tv);
-		tv.tv_sec = options_get_number(&s->options, "lock-after-time");
+		tv.tv_sec = options_get_number(s->options, "lock-after-time");
 		if (tv.tv_sec != 0)
 			evtimer_add(&s->lock_timer, &tv);
 	}
@@ -313,11 +348,11 @@ session_previous_session(struct session *s)
 /* Create a new window on a session. */
 struct winlink *
 session_new(struct session *s, const char *name, int argc, char **argv,
-    const char *path, int cwd, int idx, char **cause)
+    const char *path, const char *cwd, int idx, char **cause)
 {
 	struct window	*w;
 	struct winlink	*wl;
-	struct environ	 env;
+	struct environ	*env;
 	const char	*shell;
 	u_int		 hlimit;
 
@@ -325,30 +360,24 @@ session_new(struct session *s, const char *name, int argc, char **argv,
 		xasprintf(cause, "index in use: %d", idx);
 		return (NULL);
 	}
+	wl->session = s;
 
-	environ_init(&env);
-	environ_copy(&global_environ, &env);
-	environ_copy(&s->environ, &env);
-	server_fill_environ(s, &env);
-
-	shell = options_get_string(&s->options, "default-shell");
+	shell = options_get_string(s->options, "default-shell");
 	if (*shell == '\0' || areshell(shell))
 		shell = _PATH_BSHELL;
 
-	hlimit = options_get_number(&s->options, "history-limit");
-	w = window_create(name, argc, argv, path, shell, cwd, &env, s->tio,
+	hlimit = options_get_number(s->options, "history-limit");
+	env = environ_for_session(s, 0);
+	w = window_create_spawn(name, argc, argv, path, shell, cwd, env, s->tio,
 	    s->sx, s->sy, hlimit, cause);
 	if (w == NULL) {
 		winlink_remove(&s->windows, wl);
-		environ_free(&env);
+		environ_free(env);
 		return (NULL);
 	}
 	winlink_set_window(wl, w);
-	notify_window_linked(s, w);
-	environ_free(&env);
-
-	if (options_get_number(&s->options, "set-remain-on-exit"))
-		options_set_number(&w->options, "remain-on-exit", 1);
+	environ_free(env);
+	notify_session_window("window-linked", s, w);
 
 	session_group_synchronize_from(s);
 	return (wl);
@@ -364,8 +393,9 @@ session_attach(struct session *s, struct window *w, int idx, char **cause)
 		xasprintf(cause, "index in use: %d", idx);
 		return (NULL);
 	}
+	wl->session = s;
 	winlink_set_window(wl, w);
-	notify_window_linked(s, w);
+	notify_session_window("window-linked", s, w);
 
 	session_group_synchronize_from(s);
 	return (wl);
@@ -376,14 +406,17 @@ int
 session_detach(struct session *s, struct winlink *wl)
 {
 	if (s->curw == wl &&
-	    session_last(s) != 0 && session_previous(s, 0) != 0)
+	    session_last(s) != 0 &&
+	    session_previous(s, 0) != 0)
 		session_next(s, 0);
 
 	wl->flags &= ~WINLINK_ALERTFLAGS;
-	notify_window_unlinked(s, wl->window);
+	notify_session_window("window-unlinked", s, wl->window);
 	winlink_stack_remove(&s->lastw, wl);
 	winlink_remove(&s->windows, wl);
+
 	session_group_synchronize_from(s);
+
 	if (RB_EMPTY(&s->windows)) {
 		session_destroy(s);
 		return (1);
@@ -397,8 +430,8 @@ session_has(struct session *s, struct window *w)
 {
 	struct winlink	*wl;
 
-	RB_FOREACH(wl, winlinks, &s->windows) {
-		if (wl->window == w)
+	TAILQ_FOREACH(wl, &w->winlinks, wentry) {
+		if (wl->session == s)
 			return (1);
 	}
 	return (0);
@@ -413,12 +446,12 @@ session_is_linked(struct session *s, struct window *w)
 {
 	struct session_group	*sg;
 
-	if ((sg = session_group_find(s)) != NULL)
+	if ((sg = session_group_contains(s)) != NULL)
 		return (w->references != session_group_count(sg));
 	return (w->references != 1);
 }
 
-struct winlink *
+static struct winlink *
 session_next_alert(struct winlink *wl)
 {
 	while (wl != NULL) {
@@ -449,7 +482,7 @@ session_next(struct session *s, int alert)
 	return (session_set_current(s, wl));
 }
 
-struct winlink *
+static struct winlink *
 session_previous_alert(struct winlink *wl)
 {
 	while (wl != NULL) {
@@ -519,17 +552,18 @@ session_set_current(struct session *s, struct winlink *wl)
 	s->curw = wl;
 	winlink_clear_flags(wl);
 	window_update_activity(wl->window);
+	notify_session("session-window-changed", s);
 	return (0);
 }
 
 /* Find the session group containing a session. */
 struct session_group *
-session_group_find(struct session *target)
+session_group_contains(struct session *target)
 {
 	struct session_group	*sg;
 	struct session		*s;
 
-	TAILQ_FOREACH(sg, &session_groups, entry) {
+	RB_FOREACH(sg, session_groups, &session_groups) {
 		TAILQ_FOREACH(s, &sg->sessions, gentry) {
 			if (s == target)
 				return (sg);
@@ -538,60 +572,58 @@ session_group_find(struct session *target)
 	return (NULL);
 }
 
-/* Find session group index. */
-u_int
-session_group_index(struct session_group *sg)
+/* Find session group by name. */
+struct session_group *
+session_group_find(const char *name)
 {
-	struct session_group   *sg2;
-	u_int			i;
+	struct session_group	sg;
 
-	i = 0;
-	TAILQ_FOREACH(sg2, &session_groups, entry) {
-		if (sg == sg2)
-			return (i);
-		i++;
-	}
-
-	fatalx("session group not found");
+	sg.name = name;
+	return (RB_FIND(session_groups, &session_groups, &sg));
 }
 
-/*
- * Add a session to the session group containing target, creating it if
- * necessary.
- */
-void
-session_group_add(struct session *target, struct session *s)
+/* Create a new session group. */
+struct session_group *
+session_group_new(const char *name)
 {
 	struct session_group	*sg;
 
-	if ((sg = session_group_find(target)) == NULL) {
-		sg = xmalloc(sizeof *sg);
-		TAILQ_INSERT_TAIL(&session_groups, sg, entry);
-		TAILQ_INIT(&sg->sessions);
-		TAILQ_INSERT_TAIL(&sg->sessions, target, gentry);
-	}
-	TAILQ_INSERT_TAIL(&sg->sessions, s, gentry);
+	if ((sg = session_group_find(name)) != NULL)
+		return (sg);
+
+	sg = xcalloc(1, sizeof *sg);
+	sg->name = xstrdup(name);
+	TAILQ_INIT(&sg->sessions);
+
+	RB_INSERT(session_groups, &session_groups, sg);
+	return (sg);
+}
+
+/* Add a session to a session group. */
+void
+session_group_add(struct session_group *sg, struct session *s)
+{
+	if (session_group_contains(s) == NULL)
+		TAILQ_INSERT_TAIL(&sg->sessions, s, gentry);
 }
 
 /* Remove a session from its group and destroy the group if empty. */
-void
+static void
 session_group_remove(struct session *s)
 {
 	struct session_group	*sg;
 
-	if ((sg = session_group_find(s)) == NULL)
+	if ((sg = session_group_contains(s)) == NULL)
 		return;
 	TAILQ_REMOVE(&sg->sessions, s, gentry);
-	if (TAILQ_NEXT(TAILQ_FIRST(&sg->sessions), gentry) == NULL)
-		TAILQ_REMOVE(&sg->sessions, TAILQ_FIRST(&sg->sessions), gentry);
 	if (TAILQ_EMPTY(&sg->sessions)) {
-		TAILQ_REMOVE(&session_groups, sg, entry);
+		RB_REMOVE(session_groups, &session_groups, sg);
 		free(sg);
 	}
 }
 
 /* Count number of sessions in session group. */
-u_int
+static u_int
 session_group_count(struct session_group *sg)
 {
 	struct session	*s;
@@ -610,7 +642,7 @@ session_group_synchronize_to(struct session *s)
 	struct session_group	*sg;
 	struct session		*target;
 
-	if ((sg = session_group_find(s)) == NULL)
+	if ((sg = session_group_contains(s)) == NULL)
 		return;
 
 	target = NULL;
@@ -618,7 +650,8 @@ session_group_synchronize_to(struct session *s)
 		if (target != s)
 			break;
 	}
-	session_group_synchronize1(target, s);
+	if (target != NULL)
+		session_group_synchronize1(target, s);
 }
 
 /* Synchronize a session group to a session. */
@@ -628,7 +661,7 @@ session_group_synchronize_from(struct session *target)
 	struct session_group	*sg;
 	struct session		*s;
 
-	if ((sg = session_group_find(target)) == NULL)
+	if ((sg = session_group_contains(target)) == NULL)
 		return;
 
 	TAILQ_FOREACH(s, &sg->sessions, gentry) {
@@ -642,7 +675,7 @@ session_group_synchronize_from(struct session *target)
  * winlinks then recreating them, then updating the current window, last window
  * stack and alerts.
  */
-void
+static void
 session_group_synchronize1(struct session *target, struct session *s)
 {
 	struct winlinks		 old_windows, *ww;
@@ -667,8 +700,9 @@ session_group_synchronize1(struct session *target, struct session *s)
 	/* Link all the windows from the target. */
 	RB_FOREACH(wl, winlinks, ww) {
 		wl2 = winlink_add(&s->windows, wl->idx);
+		wl2->session = s;
 		winlink_set_window(wl2, wl->window);
-		notify_window_linked(s, wl2->window);
+		notify_session_window("window-linked", s, wl2->window);
 		wl2->flags |= wl->flags & WINLINK_ALERTFLAGS;
 	}
 
@@ -692,7 +726,7 @@ session_group_synchronize1(struct session *target, struct session *s)
 		wl = RB_ROOT(&old_windows);
 		wl2 = winlink_find_by_window_id(&s->windows, wl->window->id);
 		if (wl2 == NULL)
-			notify_window_unlinked(s, wl->window);
+			notify_session_window("window-unlinked", s, wl->window);
 		winlink_remove(&old_windows, wl);
 	}
 }
@@ -711,12 +745,13 @@ session_renumber_windows(struct session *s)
 	RB_INIT(&s->windows);
 
 	/* Start renumbering from the base-index if it's set. */
-	new_idx = options_get_number(&s->options, "base-index");
+	new_idx = options_get_number(s->options, "base-index");
 	new_curw_idx = 0;
 
 	/* Go through the winlinks and assign new indexes. */
 	RB_FOREACH(wl, winlinks, &old_wins) {
 		wl_new = winlink_add(&s->windows, new_idx);
+		wl_new->session = s;
 		winlink_set_window(wl_new, wl->window);
 		wl_new->flags |= wl->flags & WINLINK_ALERTFLAGS;
 
